from __future__ import print_function, unicode_literals, absolute_import

import sys

success = False
in_ironpython = "IronPython" in sys.version
from . import winconstants

if in_ironpython:
    try:
        from .ironpython_keysyms import *
        success = True
    except ImportError as x:
        raise
else:
    try:
        from .keysyms import *
        success = True
    except ImportError as x:
        pass
    
if not success:
    raise ImportError("Could not import keysym for local pythonversion", x)


from __future__ import print_function, unicode_literals, absolute_import

class Event(object):
    '''Represent events from the console.'''
    def __init__(self, console, input):
        pass

    def __repr__(self):
        '''Display an event for debugging.'''
        if self.type in ['KeyPress', 'KeyRelease']:
            chr = self.char
            if ord(chr) < ord("A"):
                chr = "?"
            s = "%s char='%s'%d keysym='%s' keycode=%d:%x state=%x keyinfo=%s" % \
                   (self.type, chr, ord(self.char), self.keysym, self.keycode, self.keycode,
                     self.state, self.keyinfo)
        elif self.type in ['Motion', 'Button']:
            s = '%s x=%d y=%d state=%x' % (self.type, self.x, self.y, self.state)
        elif self.type == 'Configure':
            s = '%s w=%d h=%d' % (self.type, self.width, self.height)
        elif self.type in ['FocusIn', 'FocusOut']:
            s = self.type
        elif self.type == 'Menu':
            s = '%s state=%x' % (self.type, self.state)
        else:
            s = 'unknown event type'
        return s


#    def __str__(self):
#        return "('%s',%s,%s,%s)"%(self.char,self.key,self.state,self.keyinfo)


from __future__ import print_function, unicode_literals, absolute_import
import glob, sys

success = False
in_ironpython = "IronPython" in sys.version

if in_ironpython:
    try:
        from .ironpython_console import *
        success = True
    except ImportError:
        raise
else:
    try:
        from .console import *
        success = True
    except ImportError:
        pass
        raise

if not success:
    raise ImportError(
            "Could not find a console implementation for your platform")



# -*- coding: utf-8 -*-
#*****************************************************************************
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@bostream.nu>
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from __future__ import print_function, unicode_literals, absolute_import

import logging
import logging.handlers
import struct, socket
from pyreadline.unicode_helper import ensure_unicode
try:
    import msvcrt
except ImportError:
    msvcrt = None
    print("problem")


port = logging.handlers.DEFAULT_TCP_LOGGING_PORT
host = 'localhost'

def check_key():
    if msvcrt is None:
        return False
    else:
        if msvcrt.kbhit():
            q = ensure_unicode(msvcrt.getch())
            return q
    return ""


singleline=False

def main():
    print("Starting TCP logserver on port:", port)
    print("Press q to quit logserver", port)
    print("Press c to clear screen", port)
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    s.bind(("", port))
    s.settimeout(1)
    while 1:
        try:
            data, addr = s.recvfrom(100000)
            print(data, end="")
        except socket.timeout:
            key = check_key().lower()
            if "q" == key:
                print("Quitting logserver")
                break
            elif "c" == key:
                print("\n" * 100)  

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
#*****************************************************************************
#       Copyright (C) 2003-2006 Gary Bishop.
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@bostream.nu>
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from __future__ import print_function, unicode_literals, absolute_import
'''Cursor control and color for the Windows console.

This was modeled after the C extension of the same name by Fredrik Lundh. 
'''

# primitive debug printing that won't interfere with the screen

import sys,os
import traceback
import re

import pyreadline.unicode_helper as unicode_helper

from pyreadline.logger import log
from pyreadline.unicode_helper import ensure_unicode, ensure_str
from pyreadline.keysyms import make_KeyPress, KeyPress
from pyreadline.console.ansi import AnsiState, AnsiWriter

try:
    import ctypes.util
    from ctypes import *
    from _ctypes import call_function
    from ctypes.wintypes import *
except ImportError:
    raise ImportError("You need ctypes to run this code")

if sys.version_info < (2, 6):
    bytes = str

def nolog(string):
    pass
    
log = nolog


# some constants we need
STD_INPUT_HANDLE = -10
STD_OUTPUT_HANDLE = -11
ENABLE_WINDOW_INPUT = 0x0008
ENABLE_MOUSE_INPUT = 0x0010
ENABLE_PROCESSED_INPUT = 0x0001
WHITE = 0x7
BLACK = 0
MENU_EVENT = 0x0008
KEY_EVENT = 0x0001
MOUSE_MOVED = 0x0001
MOUSE_EVENT = 0x0002
WINDOW_BUFFER_SIZE_EVENT = 0x0004
FOCUS_EVENT = 0x0010
MENU_EVENT = 0x0008
VK_SHIFT = 0x10
VK_CONTROL = 0x11
VK_MENU = 0x12
GENERIC_READ = int(0x80000000)
GENERIC_WRITE = 0x40000000

# Windows structures we'll need later
class COORD(Structure):
    _fields_ = [("X", c_short),
                ("Y", c_short)]

class SMALL_RECT(Structure):
    _fields_ = [("Left", c_short),
                ("Top", c_short),
                ("Right", c_short),
                ("Bottom", c_short)]

class CONSOLE_SCREEN_BUFFER_INFO(Structure):
    _fields_ = [("dwSize", COORD),
                ("dwCursorPosition", COORD),
                ("wAttributes", c_short),
                ("srWindow", SMALL_RECT),
                ("dwMaximumWindowSize", COORD)]

class CHAR_UNION(Union):
    _fields_ = [("UnicodeChar", c_wchar),
                ("AsciiChar", c_char)]

class CHAR_INFO(Structure):
    _fields_ = [("Char", CHAR_UNION),
                ("Attributes", c_short)]

class KEY_EVENT_RECORD(Structure):
    _fields_ = [("bKeyDown", c_byte),
                ("pad2", c_byte),
                ('pad1', c_short),
                ("wRepeatCount", c_short),
                ("wVirtualKeyCode", c_short),
                ("wVirtualScanCode", c_short),
                ("uChar", CHAR_UNION),
                ("dwControlKeyState", c_int)]

class MOUSE_EVENT_RECORD(Structure):
    _fields_ = [("dwMousePosition", COORD),
                ("dwButtonState", c_int),
                ("dwControlKeyState", c_int),
                ("dwEventFlags", c_int)]

class WINDOW_BUFFER_SIZE_RECORD(Structure):
    _fields_ = [("dwSize", COORD)]

class MENU_EVENT_RECORD(Structure):
    _fields_ = [("dwCommandId", c_uint)]

class FOCUS_EVENT_RECORD(Structure):
    _fields_ = [("bSetFocus", c_byte)]

class INPUT_UNION(Union):
    _fields_ = [("KeyEvent", KEY_EVENT_RECORD),
                ("MouseEvent", MOUSE_EVENT_RECORD),
                ("WindowBufferSizeEvent", WINDOW_BUFFER_SIZE_RECORD),
                ("MenuEvent", MENU_EVENT_RECORD),
                ("FocusEvent", FOCUS_EVENT_RECORD)]

class INPUT_RECORD(Structure):
    _fields_ = [("EventType", c_short),
                ("Event", INPUT_UNION)]

class CONSOLE_CURSOR_INFO(Structure):
    _fields_ = [("dwSize", c_int),
                ("bVisible", c_byte)]


# I didn't want to have to individually import these so I made a list, they are
# added to the Console class later in this file.

funcs = [
    'AllocConsole',
    'CreateConsoleScreenBuffer',
    'FillConsoleOutputAttribute',
    'FillConsoleOutputCharacterW',
    'FreeConsole',
    'GetConsoleCursorInfo',
    'GetConsoleMode',
    'GetConsoleScreenBufferInfo',
    'GetConsoleTitleW',
    'GetProcAddress',
    'GetStdHandle',
    'PeekConsoleInputW',
    'ReadConsoleInputW',
    'ScrollConsoleScreenBufferW',
    'SetConsoleActiveScreenBuffer',
    'SetConsoleCursorInfo',
    'SetConsoleCursorPosition',
    'SetConsoleMode',
    'SetConsoleScreenBufferSize',
    'SetConsoleTextAttribute',
    'SetConsoleTitleW',
    'SetConsoleWindowInfo',
    'WriteConsoleW',
    'WriteConsoleOutputCharacterW',
    'WriteFile',
    ]

# I don't want events for these keys, they are just a bother for my application
key_modifiers = { VK_SHIFT : 1,
                  VK_CONTROL : 1,
                  VK_MENU : 1, # alt key
                  0x5b : 1, # windows key
                 }

def split_block(text, size=1000):
    return [text[start:start + size] for start in range(0, len(text), size)]



class Console(object):
    '''Console driver for Windows.

    '''

    def __init__(self, newbuffer=0):
        '''Initialize the Console object.

        newbuffer=1 will allocate a new buffer so the old content will be restored
        on exit.
        '''
        #Do I need the following line? It causes a console to be created whenever
        #readline is imported into a pythonw application which seems wrong. Things
        #seem to work without it...
        #self.AllocConsole()

        if newbuffer:
            self.hout = self.CreateConsoleScreenBuffer(
                             GENERIC_READ | GENERIC_WRITE,
                             0, None, 1, None)
            self.SetConsoleActiveScreenBuffer(self.hout)
        else:
            self.hout = self.GetStdHandle(STD_OUTPUT_HANDLE)

        self.hin = self.GetStdHandle(STD_INPUT_HANDLE)
        self.inmode = DWORD(0)
        self.GetConsoleMode(self.hin, byref(self.inmode))
        self.SetConsoleMode(self.hin, 0xf)
        info = CONSOLE_SCREEN_BUFFER_INFO()
        self.GetConsoleScreenBufferInfo(self.hout, byref(info))
        self.attr = info.wAttributes
        self.saveattr = info.wAttributes # remember the initial colors
        self.defaultstate = AnsiState()
        self.defaultstate.winattr = info.wAttributes
        self.ansiwriter = AnsiWriter(self.defaultstate)
        
        background = self.attr & 0xf0
        for escape in self.escape_to_color:
            if self.escape_to_color[escape] is not None:
                self.escape_to_color[escape] |= background
        log('initial attr=%x' % self.attr)
        self.softspace = 0 # this is for using it as a file-like object
        self.serial = 0

        self.pythondll = ctypes.pythonapi
        self.inputHookPtr = \
            c_void_p.from_address(addressof(self.pythondll.PyOS_InputHook)).value

        if sys.version_info[:2] > (3, 4): 
            self.pythondll.PyMem_RawMalloc.restype = c_size_t
            self.pythondll.PyMem_Malloc.argtypes = [c_size_t]
            setattr(Console, 'PyMem_Malloc', self.pythondll.PyMem_RawMalloc)
        else:
            self.pythondll.PyMem_Malloc.restype = c_size_t
            self.pythondll.PyMem_Malloc.argtypes = [c_size_t]
            setattr(Console, 'PyMem_Malloc', self.pythondll.PyMem_Malloc)



        
    def __del__(self):
        '''Cleanup the console when finished.'''
        # I don't think this ever gets called
        self.SetConsoleTextAttribute(self.hout, self.saveattr)
        self.SetConsoleMode(self.hin, self.inmode)
        self.FreeConsole()

    def _get_top_bot(self):
        info = CONSOLE_SCREEN_BUFFER_INFO()
        self.GetConsoleScreenBufferInfo(self.hout, byref(info))
        rect = info.srWindow
        top = rect.Top 
        bot = rect.Bottom 
        return top,bot

    def fixcoord(self, x, y):
        '''Return a long with x and y packed inside, 
        also handle negative x and y.'''
        if x < 0 or y < 0:
            info = CONSOLE_SCREEN_BUFFER_INFO()
            self.GetConsoleScreenBufferInfo(self.hout, byref(info))
            if x < 0:
                x = info.srWindow.Right - x
                y = info.srWindow.Bottom + y

        # this is a hack! ctypes won't pass structures but COORD is 
        # just like a long, so this works.
        return c_int(y << 16 | x)

    def pos(self, x=None, y=None):
        '''Move or query the window cursor.'''
        if x is None:
            info = CONSOLE_SCREEN_BUFFER_INFO()
            self.GetConsoleScreenBufferInfo(self.hout, byref(info))
            return (info.dwCursorPosition.X, info.dwCursorPosition.Y)
        else:
            return self.SetConsoleCursorPosition(self.hout, 
                                                 self.fixcoord(x, y))

    def home(self):
        '''Move to home.'''
        self.pos(0, 0)

# Map ANSI color escape sequences into Windows Console Attributes

    terminal_escape = re.compile('(\001?\033\\[[0-9;]+m\002?)')
    escape_parts = re.compile('\001?\033\\[([0-9;]+)m\002?')
    escape_to_color = { '0;30': 0x0,             #black
                        '0;31': 0x4,             #red
                        '0;32': 0x2,             #green
                        '0;33': 0x4+0x2,         #brown?
                        '0;34': 0x1,             #blue
                        '0;35': 0x1+0x4,         #purple
                        '0;36': 0x2+0x4,         #cyan
                        '0;37': 0x1+0x2+0x4,     #grey
                        '1;30': 0x1+0x2+0x4,     #dark gray
                        '1;31': 0x4+0x8,         #red
                        '1;32': 0x2+0x8,         #light green
                        '1;33': 0x4+0x2+0x8,     #yellow
                        '1;34': 0x1+0x8,         #light blue
                        '1;35': 0x1+0x4+0x8,     #light purple
                        '1;36': 0x1+0x2+0x8,     #light cyan
                        '1;37': 0x1+0x2+0x4+0x8, #white
                        '0': None,
                       }

    # This pattern should match all characters that change the cursor position differently
    # than a normal character.
    motion_char_re = re.compile('([\n\r\t\010\007])')

    def write_scrolling(self, text, attr=None):
        '''write text at current cursor position while watching for scrolling.

        If the window scrolls because you are at the bottom of the screen
        buffer, all positions that you are storing will be shifted by the
        scroll amount. For example, I remember the cursor position of the
        prompt so that I can redraw the line but if the window scrolls,
        the remembered position is off.

        This variant of write tries to keep track of the cursor position
        so that it will know when the screen buffer is scrolled. It
        returns the number of lines that the buffer scrolled.

        '''
        text = ensure_unicode(text)
        x, y = self.pos()
        w, h = self.size()
        scroll = 0 # the result
        # split the string into ordinary characters and funny characters
        chunks = self.motion_char_re.split(text)
        for chunk in chunks:
            n = self.write_color(chunk, attr)
            if len(chunk) == 1: # the funny characters will be alone
                if chunk[0] == '\n': # newline
                    x = 0
                    y += 1
                elif chunk[0] == '\r': # carriage return
                    x = 0
                elif chunk[0] == '\t': # tab
                    x = 8 * (int(x / 8) + 1)
                    if x > w: # newline
                        x -= w
                        y += 1
                elif chunk[0] == '\007': # bell
                    pass
                elif chunk[0] == '\010':
                    x -= 1
                    if x < 0:
                        y -= 1 # backed up 1 line
                else: # ordinary character
                    x += 1
                if x == w: # wrap
                    x = 0
                    y += 1
                if y == h: # scroll
                    scroll += 1
                    y = h - 1
            else: # chunk of ordinary characters
                x += n
                l = int(x / w) # lines we advanced
                x = x % w # new x value
                y += l
                if y >= h: # scroll
                    scroll += y - h + 1
                    y = h - 1
        return scroll

    def write_color(self, text, attr=None):
        text = ensure_unicode(text)
        n, res= self.ansiwriter.write_color(text, attr)
        junk = DWORD(0)
        for attr,chunk in res:
            log("console.attr:%s"%(attr))
            log("console.chunk:%s"%(chunk))
            self.SetConsoleTextAttribute(self.hout, attr.winattr)
            for short_chunk in split_block(chunk):
                self.WriteConsoleW(self.hout, short_chunk, 
                                   len(short_chunk), byref(junk), None)
        return n

    def write_plain(self, text, attr=None):
        '''write text at current cursor position.'''
        text = ensure_unicode(text)
        log('write("%s", %s)' %(text, attr))
        if attr is None:
            attr = self.attr
        junk = DWORD(0)
        self.SetConsoleTextAttribute(self.hout, attr)
        for short_chunk in split_block(chunk):
            self.WriteConsoleW(self.hout, ensure_unicode(short_chunk), 
                               len(short_chunk), byref(junk), None)
        return len(text)

    #This function must be used to ensure functioning with EMACS
    #Emacs sets the EMACS environment variable
    if "EMACS" in os.environ:
        def write_color(self, text, attr=None):
            text = ensure_str(text)
            junk = DWORD(0)
            self.WriteFile(self.hout, text, len(text), byref(junk), None)
            return len(text)
        write_plain = write_color

    # make this class look like a file object
    def write(self, text):
        text = ensure_unicode(text)
        log('write("%s")' % text)
        return self.write_color(text)

    #write = write_scrolling

    def isatty(self):
        return True

    def flush(self):
        pass

    def page(self, attr=None, fill=' '):
        '''Fill the entire screen.'''
        if attr is None:
            attr = self.attr
        if len(fill) != 1:
            raise ValueError
        info = CONSOLE_SCREEN_BUFFER_INFO()
        self.GetConsoleScreenBufferInfo(self.hout, byref(info))
        if info.dwCursorPosition.X != 0 or info.dwCursorPosition.Y != 0:
            self.SetConsoleCursorPosition(self.hout, self.fixcoord(0, 0))

        w = info.dwSize.X
        n = DWORD(0)
        for y in range(info.dwSize.Y):
            self.FillConsoleOutputAttribute(self.hout, attr, 
                                            w, self.fixcoord(0, y), byref(n))
            self.FillConsoleOutputCharacterW(self.hout, ord(fill[0]), 
                                             w, self.fixcoord(0, y), byref(n))

        self.attr = attr

    def text(self, x, y, text, attr=None):
        '''Write text at the given position.'''
        if attr is None:
            attr = self.attr

        pos = self.fixcoord(x, y)
        n = DWORD(0)
        self.WriteConsoleOutputCharacterW(self.hout, text, 
                                          len(text), pos, byref(n))
        self.FillConsoleOutputAttribute(self.hout, attr, n, pos, byref(n))

    def clear_to_end_of_window(self):
        top, bot = self._get_top_bot()
        pos = self.pos()
        w, h = self.size()
        self.rectangle( (pos[0], pos[1], w, pos[1] + 1))
        if pos[1] < bot:
            self.rectangle((0, pos[1] + 1, w, bot + 1))

    def rectangle(self, rect, attr=None, fill=' '):
        '''Fill Rectangle.'''
        x0, y0, x1, y1 = rect
        n = DWORD(0)
        if attr is None:
            attr = self.attr
        for y in range(y0, y1):
            pos = self.fixcoord(x0, y)
            self.FillConsoleOutputAttribute(self.hout, attr, x1 - x0, 
                                            pos, byref(n))
            self.FillConsoleOutputCharacterW(self.hout, ord(fill[0]), x1 - x0,
                                             pos, byref(n))

    def scroll(self, rect, dx, dy, attr=None, fill=' '):
        '''Scroll a rectangle.'''
        if attr is None:
            attr = self.attr
        x0, y0, x1, y1 = rect
        source = SMALL_RECT(x0, y0, x1 - 1, y1 - 1)
        dest = self.fixcoord(x0 + dx, y0 + dy)
        style = CHAR_INFO()
        style.Char.AsciiChar = ensure_str(fill[0])
        style.Attributes = attr

        return self.ScrollConsoleScreenBufferW(self.hout, byref(source), 
                                          byref(source), dest, byref(style))

    def scroll_window(self, lines):
        '''Scroll the window by the indicated number of lines.'''
        info = CONSOLE_SCREEN_BUFFER_INFO()
        self.GetConsoleScreenBufferInfo(self.hout, byref(info))
        rect = info.srWindow
        log('sw: rtop=%d rbot=%d' % (rect.Top, rect.Bottom))
        top = rect.Top + lines
        bot = rect.Bottom + lines
        h = bot - top
        maxbot = info.dwSize.Y-1
        if top < 0:
            top = 0
            bot = h
        if bot > maxbot:
            bot = maxbot
            top = bot - h

        nrect = SMALL_RECT()
        nrect.Top = top
        nrect.Bottom = bot
        nrect.Left = rect.Left
        nrect.Right = rect.Right
        log('sn: top=%d bot=%d' % (top, bot))
        r=self.SetConsoleWindowInfo(self.hout, True, byref(nrect))
        log('r=%d' % r)

    def get(self):
        '''Get next event from queue.'''
        inputHookFunc = c_void_p.from_address(self.inputHookPtr).value

        Cevent = INPUT_RECORD()
        count = DWORD(0)
        while 1:
            if inputHookFunc:
                call_function(inputHookFunc, ())
            status = self.ReadConsoleInputW(self.hin, 
                                        byref(Cevent), 1, byref(count))
            if status and count.value == 1:
                e = event(self, Cevent)
                return e

    def getkeypress(self):
        '''Return next key press event from the queue, ignoring others.'''
        while 1:
            e = self.get()
            if e.type == 'KeyPress' and e.keycode not in key_modifiers:
                log("console.getkeypress %s"%e)
                if e.keyinfo.keyname == 'next':
                    self.scroll_window(12)
                elif e.keyinfo.keyname == 'prior':
                    self.scroll_window(-12)
                else:
                    return e
            elif ((e.type == 'KeyRelease') and 
                  (e.keyinfo == KeyPress('S', False, True, False, 'S'))):
                log("getKeypress:%s,%s,%s"%(e.keyinfo, e.keycode, e.type))
                return e
            
                
    def getchar(self):
        '''Get next character from queue.'''

        Cevent = INPUT_RECORD()
        count = DWORD(0)
        while 1:
            status = self.ReadConsoleInputW(self.hin, 
                                            byref(Cevent), 1, byref(count))
            if (status and 
                (count.value == 1) and 
                (Cevent.EventType == 1) and
                Cevent.Event.KeyEvent.bKeyDown):
                sym = keysym(Cevent.Event.KeyEvent.wVirtualKeyCode)
                if len(sym) == 0:
                    sym = Cevent.Event.KeyEvent.uChar.AsciiChar
                return sym

    def peek(self):
        '''Check event queue.'''
        Cevent = INPUT_RECORD()
        count = DWORD(0)
        status = self.PeekConsoleInputW(self.hin, 
                                byref(Cevent), 1, byref(count))
        if status and count == 1:
            return event(self, Cevent)

    def title(self, txt=None):
        '''Set/get title.'''
        if txt:
            self.SetConsoleTitleW(txt)
        else:
            buffer = create_unicode_buffer(200)
            n = self.GetConsoleTitleW(buffer, 200)
            if n > 0:
                return buffer.value[:n]

    def size(self, width=None, height=None):
        '''Set/get window size.'''
        info = CONSOLE_SCREEN_BUFFER_INFO()
        status = self.GetConsoleScreenBufferInfo(self.hout, byref(info))
        if not status:
            return None
        if width is not None and height is not None:
            wmin = info.srWindow.Right - info.srWindow.Left + 1
            hmin = info.srWindow.Bottom - info.srWindow.Top + 1
            #print wmin, hmin
            width = max(width, wmin)
            height = max(height, hmin)
            #print width, height
            self.SetConsoleScreenBufferSize(self.hout, 
                                            self.fixcoord(width, height))
        else:
            return (info.dwSize.X, info.dwSize.Y)

    def cursor(self, visible=None, size=None):
        '''Set cursor on or off.'''
        info = CONSOLE_CURSOR_INFO()
        if self.GetConsoleCursorInfo(self.hout, byref(info)):
            if visible is not None:
                info.bVisible = visible
            if size is not None:
                info.dwSize = size
            self.SetConsoleCursorInfo(self.hout, byref(info))

    def bell(self):
        self.write('\007')

    def next_serial(self):
        '''Get next event serial number.'''
        self.serial += 1
        return self.serial

# add the functions from the dll to the class
for func in funcs:
    setattr(Console, func, getattr(windll.kernel32, func))


_strncpy = ctypes.windll.kernel32.lstrcpynA
_strncpy.restype = c_char_p
_strncpy.argtypes = [c_char_p, c_char_p, c_size_t]


LPVOID = c_void_p
LPCVOID = c_void_p
FARPROC = c_void_p
LPDWORD = POINTER(DWORD)

Console.AllocConsole.restype = BOOL
Console.AllocConsole.argtypes = [] #void
Console.CreateConsoleScreenBuffer.restype = HANDLE
Console.CreateConsoleScreenBuffer.argtypes = [DWORD, DWORD, c_void_p, DWORD, LPVOID] #DWORD, DWORD, SECURITY_ATTRIBUTES*, DWORD, LPVOID  
Console.FillConsoleOutputAttribute.restype = BOOL 
Console.FillConsoleOutputAttribute.argtypes = [HANDLE, WORD, DWORD, c_int, LPDWORD] #HANDLE, WORD, DWORD, COORD, LPDWORD
Console.FillConsoleOutputCharacterW.restype = BOOL
Console.FillConsoleOutputCharacterW.argtypes = [HANDLE, c_ushort, DWORD, c_int, LPDWORD] #HANDLE, TCHAR, DWORD, COORD, LPDWORD
Console.FreeConsole.restype = BOOL
Console.FreeConsole.argtypes = [] #void
Console.GetConsoleCursorInfo.restype = BOOL
Console.GetConsoleCursorInfo.argtypes = [HANDLE, c_void_p] #HANDLE, PCONSOLE_CURSOR_INFO  
Console.GetConsoleMode.restype = BOOL
Console.GetConsoleMode.argtypes = [HANDLE, LPDWORD] #HANDLE, LPDWORD  
Console.GetConsoleScreenBufferInfo.restype = BOOL
Console.GetConsoleScreenBufferInfo.argtypes = [HANDLE, c_void_p] #HANDLE, PCONSOLE_SCREEN_BUFFER_INFO 
Console.GetConsoleTitleW.restype = DWORD
Console.GetConsoleTitleW.argtypes = [c_wchar_p, DWORD] #LPTSTR , DWORD
Console.GetProcAddress.restype = FARPROC
Console.GetProcAddress.argtypes = [HMODULE, c_char_p] #HMODULE , LPCSTR 
Console.GetStdHandle.restype = HANDLE
Console.GetStdHandle.argtypes = [DWORD]
Console.PeekConsoleInputW.restype = BOOL
Console.PeekConsoleInputW.argtypes = [HANDLE, c_void_p, DWORD, LPDWORD] #HANDLE, PINPUT_RECORD, DWORD, LPDWORD
Console.ReadConsoleInputW.restype = BOOL
Console.ReadConsoleInputW.argtypes = [HANDLE, c_void_p, DWORD, LPDWORD] #HANDLE, PINPUT_RECORD, DWORD, LPDWORD
Console.ScrollConsoleScreenBufferW.restype = BOOL
Console.ScrollConsoleScreenBufferW.argtypes = [HANDLE, c_void_p, c_void_p, c_int, c_void_p] #HANDLE, SMALL_RECT*, SMALL_RECT*, COORD, LPDWORD
Console.SetConsoleActiveScreenBuffer.restype = BOOL
Console.SetConsoleActiveScreenBuffer.argtypes = [HANDLE] #HANDLE
Console.SetConsoleCursorInfo.restype = BOOL
Console.SetConsoleCursorInfo.argtypes = [HANDLE, c_void_p] #HANDLE, CONSOLE_CURSOR_INFO* 
Console.SetConsoleCursorPosition.restype = BOOL
Console.SetConsoleCursorPosition.argtypes = [HANDLE, c_int] #HANDLE, COORD 
Console.SetConsoleMode.restype = BOOL
Console.SetConsoleMode.argtypes = [HANDLE, DWORD] #HANDLE, DWORD
Console.SetConsoleScreenBufferSize.restype = BOOL
Console.SetConsoleScreenBufferSize.argtypes = [HANDLE, c_int] #HANDLE, COORD 
Console.SetConsoleTextAttribute.restype = BOOL
Console.SetConsoleTextAttribute.argtypes = [HANDLE, WORD] #HANDLE, WORD 
Console.SetConsoleTitleW.restype = BOOL
Console.SetConsoleTitleW.argtypes = [c_wchar_p] #LPCTSTR
Console.SetConsoleWindowInfo.restype = BOOL
Console.SetConsoleWindowInfo.argtypes = [HANDLE, BOOL, c_void_p] #HANDLE, BOOL, SMALL_RECT*
Console.WriteConsoleW.restype = BOOL
Console.WriteConsoleW.argtypes = [HANDLE, c_void_p, DWORD, LPDWORD, LPVOID] #HANDLE, VOID*, DWORD, LPDWORD, LPVOID
Console.WriteConsoleOutputCharacterW.restype = BOOL
Console.WriteConsoleOutputCharacterW.argtypes = [HANDLE, c_wchar_p, DWORD, c_int, LPDWORD] #HANDLE, LPCTSTR, DWORD, COORD, LPDWORD
Console.WriteFile.restype = BOOL
Console.WriteFile.argtypes = [HANDLE, LPCVOID, DWORD, LPDWORD, c_void_p] #HANDLE, LPCVOID , DWORD, LPDWORD , LPOVERLAPPED 



from .event import Event

VkKeyScan = windll.user32.VkKeyScanA


class event(Event):
    '''Represent events from the console.'''
    def __init__(self, console, input):
        '''Initialize an event from the Windows input structure.'''
        self.type = '??'
        self.serial = console.next_serial()
        self.width = 0
        self.height = 0
        self.x = 0
        self.y = 0
        self.char = ''
        self.keycode = 0
        self.keysym = '??'
        self.keyinfo = None # a tuple with (control, meta, shift, keycode) for dispatch
        self.width = None
        
        if input.EventType == KEY_EVENT:
            if input.Event.KeyEvent.bKeyDown:
                self.type = "KeyPress"
            else:
                self.type = "KeyRelease"
            self.char = input.Event.KeyEvent.uChar.UnicodeChar
            self.keycode = input.Event.KeyEvent.wVirtualKeyCode
            self.state = input.Event.KeyEvent.dwControlKeyState
            self.keyinfo = make_KeyPress(self.char,self.state,self.keycode)

        elif input.EventType == MOUSE_EVENT:
            if input.Event.MouseEvent.dwEventFlags & MOUSE_MOVED:
                self.type = "Motion"
            else:
                self.type = "Button"
            self.x = input.Event.MouseEvent.dwMousePosition.X
            self.y = input.Event.MouseEvent.dwMousePosition.Y
            self.state = input.Event.MouseEvent.dwButtonState
        elif input.EventType == WINDOW_BUFFER_SIZE_EVENT:
            self.type = "Configure"
            self.width = input.Event.WindowBufferSizeEvent.dwSize.X
            self.height = input.Event.WindowBufferSizeEvent.dwSize.Y
        elif input.EventType == FOCUS_EVENT:
            if input.Event.FocusEvent.bSetFocus:
                self.type = "FocusIn"
            else:
                self.type = "FocusOut"
        elif input.EventType == MENU_EVENT:
            self.type = "Menu"
            self.state = input.Event.MenuEvent.dwCommandId


def getconsole(buffer=1):
        """Get a console handle.

        If buffer is non-zero, a new console buffer is allocated and
        installed.  Otherwise, this returns a handle to the current
        console buffer"""

        c = Console(buffer)

        return c

# The following code uses ctypes to allow a Python callable to
# substitute for GNU readline within the Python interpreter. Calling
# raw_input or other functions that do input, inside your callable
# might be a bad idea, then again, it might work.

# The Python callable can raise EOFError or KeyboardInterrupt and
# these will be translated into the appropriate outputs from readline
# so that they will then be translated back!

# If the Python callable raises any other exception, a traceback will
# be printed and readline will appear to return an empty line.

# I use ctypes to create a C-callable from a Python wrapper that
# handles the exceptions and gets the result into the right form.

# the type for our C-callable wrapper
HOOKFUNC23 = CFUNCTYPE(c_char_p, c_void_p, c_void_p, c_char_p)

readline_hook = None # the python hook goes here
readline_ref = None  # reference to the c-callable to keep it alive

def hook_wrapper_23(stdin, stdout, prompt):
    '''Wrap a Python readline so it behaves like GNU readline.'''
    try:
        # call the Python hook
        res = ensure_str(readline_hook(prompt))
        # make sure it returned the right sort of thing
        if res and not isinstance(res, bytes):
            raise TypeError('readline must return a string.')
    except KeyboardInterrupt:
        # GNU readline returns 0 on keyboard interrupt
        return 0
    except EOFError:
        # It returns an empty string on EOF
        res = ensure_str('')
    except:
        print('Readline internal error', file=sys.stderr)
        traceback.print_exc()
        res = ensure_str('\n')
    # we have to make a copy because the caller expects to free the result
    n = len(res)
    p = Console.PyMem_Malloc(n + 1)
    _strncpy(cast(p, c_char_p), res, n + 1)
    return p


def install_readline(hook):
    '''Set up things for the interpreter to call 
    our function like GNU readline.'''
    global readline_hook, readline_ref
    # save the hook so the wrapper can call it
    readline_hook = hook
    # get the address of PyOS_ReadlineFunctionPointer so we can update it
    PyOS_RFP = c_void_p.from_address(Console.GetProcAddress(sys.dllhandle,
                           "PyOS_ReadlineFunctionPointer".encode('ascii')))
    # save a reference to the generated C-callable so it doesn't go away
    readline_ref = HOOKFUNC23(hook_wrapper_23)
    # get the address of the function
    func_start = c_void_p.from_address(addressof(readline_ref)).value
    # write the function address into PyOS_ReadlineFunctionPointer
    PyOS_RFP.value = func_start


if __name__ == '__main__':
    import time, sys

    
    def p(char):
        return chr(VkKeyScan(ord(char)) & 0xff)

    c = Console(0)
    sys.stdout = c
    sys.stderr = c
    c.page()
    print(p("d"), p("D"))
    c.pos(5, 10)
    c.write('hi there')
    print('some printed output')
    for i in range(10):
        q = c.getkeypress()
        print(q)
    del c



#This file contains constants that are normally found in win32all
#But included here to avoid the dependency
from __future__ import print_function, unicode_literals, absolute_import

VK_LBUTTON=1
VK_RBUTTON=2
VK_CANCEL=3
VK_MBUTTON=4
VK_XBUTTON1=5
VK_XBUTTON2=6
VK_BACK=8
VK_TAB=9
VK_CLEAR=12
VK_RETURN=13
VK_SHIFT=16
VK_CONTROL=17
VK_MENU=18
VK_PAUSE=19
VK_CAPITAL=20
VK_KANA=0x15
VK_HANGEUL=0x15
VK_HANGUL=0x15
VK_JUNJA=0x17
VK_FINAL=0x18
VK_HANJA=0x19
VK_KANJI=0x19
VK_ESCAPE=0x1B
VK_CONVERT=0x1C
VK_NONCONVERT=0x1D
VK_ACCEPT=0x1E
VK_MODECHANGE=0x1F
VK_SPACE=32
VK_PRIOR=33
VK_NEXT=34
VK_END=35
VK_HOME=36
VK_LEFT=37
VK_UP=38
VK_RIGHT=39
VK_DOWN=40
VK_SELECT=41
VK_PRINT=42
VK_EXECUTE=43
VK_SNAPSHOT=44
VK_INSERT=45
VK_DELETE=46
VK_HELP=47
VK_LWIN=0x5B
VK_RWIN=0x5C
VK_APPS=0x5D
VK_SLEEP=0x5F
VK_NUMPAD0=0x60
VK_NUMPAD1=0x61
VK_NUMPAD2=0x62
VK_NUMPAD3=0x63
VK_NUMPAD4=0x64
VK_NUMPAD5=0x65
VK_NUMPAD6=0x66
VK_NUMPAD7=0x67
VK_NUMPAD8=0x68
VK_NUMPAD9=0x69
VK_MULTIPLY=0x6A
VK_ADD=0x6B
VK_SEPARATOR=0x6C
VK_SUBTRACT=0x6D
VK_DECIMAL=0x6E
VK_DIVIDE=0x6F
VK_F1=0x70
VK_F2=0x71
VK_F3=0x72
VK_F4=0x73
VK_F5=0x74
VK_F6=0x75
VK_F7=0x76
VK_F8=0x77
VK_F9=0x78
VK_F10=0x79
VK_F11=0x7A
VK_F12=0x7B
VK_F13=0x7C
VK_F14=0x7D
VK_F15=0x7E
VK_F16=0x7F
VK_F17=0x80
VK_F18=0x81
VK_F19=0x82
VK_F20=0x83
VK_F21=0x84
VK_F22=0x85
VK_F23=0x86
VK_F24=0x87
VK_NUMLOCK=0x90
VK_SCROLL=0x91
VK_LSHIFT=0xA0
VK_RSHIFT=0xA1
VK_LCONTROL=0xA2
VK_RCONTROL=0xA3
VK_LMENU=0xA4
VK_RMENU=0xA5
VK_BROWSER_BACK=0xA6
VK_BROWSER_FORWARD=0xA7
VK_BROWSER_REFRESH=0xA8
VK_BROWSER_STOP=0xA9
VK_BROWSER_SEARCH=0xAA
VK_BROWSER_FAVORITES=0xAB
VK_BROWSER_HOME=0xAC
VK_VOLUME_MUTE=0xAD
VK_VOLUME_DOWN=0xAE
VK_VOLUME_UP=0xAF
VK_MEDIA_NEXT_TRACK=0xB0
VK_MEDIA_PREV_TRACK=0xB1
VK_MEDIA_STOP=0xB2
VK_MEDIA_PLAY_PAUSE=0xB3
VK_LAUNCH_MAIL=0xB4
VK_LAUNCH_MEDIA_SELECT=0xB5
VK_LAUNCH_APP1=0xB6
VK_LAUNCH_APP2=0xB7
VK_OEM_1=0xBA
VK_OEM_PLUS=0xBB
VK_OEM_COMMA=0xBC
VK_OEM_MINUS=0xBD
VK_OEM_PERIOD=0xBE
VK_OEM_2=0xBF
VK_OEM_3=0xC0
VK_OEM_4=0xDB
VK_OEM_5=0xDC
VK_OEM_6=0xDD
VK_OEM_7=0xDE
VK_OEM_8=0xDF
VK_OEM_102=0xE2
VK_PROCESSKEY=0xE5
VK_PACKET=0xE7
VK_ATTN=0xF6
VK_CRSEL=0xF7
VK_EXSEL=0xF8
VK_EREOF=0xF9
VK_PLAY=0xFA
VK_ZOOM=0xFB
VK_NONAME=0xFC
VK_PA1=0xFD
VK_OEM_CLEAR=0xFE

CF_TEXT=1
CF_BITMAP=2
CF_METAFILEPICT=3
CF_SYLK=4
CF_DIF=5
CF_TIFF=6
CF_OEMTEXT=7
CF_DIB=8
CF_PALETTE=9
CF_PENDATA=10
CF_RIFF=11
CF_WAVE=12
CF_UNICODETEXT=13
CF_ENHMETAFILE=14
CF_HDROP=15
CF_LOCALE=16
CF_MAX=17
CF_OWNERDISPLAY=128
CF_DSPTEXT=129
CF_DSPBITMAP=130
CF_DSPMETAFILEPICT=131
CF_DSPENHMETAFILE=142
CF_PRIVATEFIRST=512
CF_PRIVATELAST=767
CF_GDIOBJFIRST=768
CF_GDIOBJLAST=1023


GPTR=64
GHND=66



# -*- coding: utf-8 -*-
#*****************************************************************************
#       Copyright (C) 2003-2006 Gary Bishop.
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@bostream.nu>
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from __future__ import print_function, unicode_literals, absolute_import
''' an attempt to implement readline for Python in Python using ctypes'''
import sys,os,re,time
from glob import glob

from . import release
from .py3k_compat import callable, execfile

import pyreadline.lineeditor.lineobj as lineobj
import pyreadline.lineeditor.history as history
import pyreadline.clipboard as clipboard
import pyreadline.console as console
import pyreadline.logger as logger 

from pyreadline.keysyms.common import make_KeyPress_from_keydescr
from pyreadline.unicode_helper import ensure_unicode, ensure_str
from .logger import log
from .modes import editingmodes
from .error import ReadlineError, GetSetError

in_ironpython = "IronPython" in sys.version
if in_ironpython:#ironpython does not provide a prompt string to readline
    import System    
    default_prompt = ">>> "
else:
    default_prompt = ""


class MockConsoleError(Exception):
    pass

class MockConsole(object):
    """object used during refactoring. Should raise errors when someone tries to use it.
    """
    def __setattr__(self, x):
        raise MockConsoleError("Should not try to get attributes from MockConsole")

    def cursor(self, size=50):
        pass

class BaseReadline(object):
    def __init__(self):
        self.allow_ctrl_c = False
        self.ctrl_c_tap_time_interval = 0.3

        self.debug = False
        self.bell_style = 'none'
        self.mark = -1
        self.console=MockConsole()
        self.disable_readline = False
        # this code needs to follow l_buffer and history creation
        self.editingmodes = [mode(self) for mode in editingmodes]
        for mode in self.editingmodes:
            mode.init_editing_mode(None)
        self.mode = self.editingmodes[0]

        self.read_inputrc()
        log("\n".join(self.mode.rl_settings_to_string()))

        self.callback = None

    def parse_and_bind(self, string):
        '''Parse and execute single line of a readline init file.'''
        try:
            log('parse_and_bind("%s")' % string)
            if string.startswith('#'):
                return
            if string.startswith('set'):
                m = re.compile(r'set\s+([-a-zA-Z0-9]+)\s+(.+)\s*$').match(string)
                if m:
                    var_name = m.group(1)
                    val = m.group(2)
                    try:
                        setattr(self.mode, var_name.replace('-','_'), val)
                    except AttributeError:
                        log('unknown var="%s" val="%s"' % (var_name, val))
                else:
                    log('bad set "%s"' % string)
                return
            m = re.compile(r'\s*(.+)\s*:\s*([-a-zA-Z]+)\s*$').match(string)
            if m:
                key = m.group(1)
                func_name = m.group(2)
                py_name = func_name.replace('-', '_')
                try:
                    func = getattr(self.mode, py_name)
                except AttributeError:
                    log('unknown func key="%s" func="%s"' % (key, func_name))
                    if self.debug:
                        print('pyreadline parse_and_bind error, unknown function to bind: "%s"' % func_name)
                    return
                self.mode._bind_key(key, func)
        except:
            log('error')
            raise

    def _set_prompt(self, prompt):
        self.mode.prompt = prompt
        
    def _get_prompt(self):
        return self.mode.prompt
    
    prompt = property(_get_prompt, _set_prompt)


    def get_line_buffer(self):
        '''Return the current contents of the line buffer.'''
        return self.mode.l_buffer.get_line_text()

    def insert_text(self, string):
        '''Insert text into the command line.'''
        self.mode.insert_text(string)
        
    def read_init_file(self, filename=None): 
        '''Parse a readline initialization file. The default filename is the last filename used.'''
        log('read_init_file("%s")' % filename)

    #History file book keeping methods (non-bindable)
    
    def add_history(self, line):
        '''Append a line to the history buffer, as if it was the last line typed.'''
        self.mode._history.add_history(line)

    def get_current_history_length(self ):
        '''Return the number of lines currently in the history.
        (This is different from get_history_length(), which returns 
        the maximum number of lines that will be written to a history file.)'''
        return self.mode._history.get_current_history_length()

    def get_history_length(self ):
        '''Return the desired length of the history file.

        Negative values imply unlimited history file size.'''
        return self.mode._history.get_history_length()

    def set_history_length(self, length): 
        '''Set the number of lines to save in the history file.

        write_history_file() uses this value to truncate the history file
        when saving. Negative values imply unlimited history file size.
        '''
        self.mode._history.set_history_length(length)

    def get_history_item(self, index): 
        '''Return the current contents of history item at index.'''
        return self.mode._history.get_history_item(index)

    def clear_history(self):
        '''Clear readline history'''
        self.mode._history.clear_history()

    def read_history_file(self, filename=None): 
        '''Load a readline history file. The default filename is ~/.history.'''
        if filename is None:
            filename = self.mode._history.history_filename
        log("read_history_file from %s"%ensure_unicode(filename))
        self.mode._history.read_history_file(filename)

    def write_history_file(self, filename=None): 
        '''Save a readline history file. The default filename is ~/.history.'''
        self.mode._history.write_history_file(filename)

    #Completer functions

    def set_completer(self, function=None): 
        '''Set or remove the completer function.

        If function is specified, it will be used as the new completer
        function; if omitted or None, any completer function already
        installed is removed. The completer function is called as
        function(text, state), for state in 0, 1, 2, ..., until it returns a
        non-string value. It should return the next possible completion
        starting with text.
        '''
        log('set_completer')
        self.mode.completer = function

    def get_completer(self): 
        '''Get the completer function. 
        '''
        log('get_completer')
        return self.mode.completer

    def get_begidx(self):
        '''Get the beginning index of the readline tab-completion scope.'''
        return self.mode.begidx

    def get_endidx(self):
        '''Get the ending index of the readline tab-completion scope.'''
        return self.mode.endidx

    def set_completer_delims(self, string):
        '''Set the readline word delimiters for tab-completion.'''
        self.mode.completer_delims = string

    def get_completer_delims(self):
        '''Get the readline word delimiters for tab-completion.'''
        if sys.version_info[0] < 3:
            return self.mode.completer_delims.encode("ascii") 
        else:
            return self.mode.completer_delims

    def set_startup_hook(self, function=None): 
        '''Set or remove the startup_hook function.

        If function is specified, it will be used as the new startup_hook
        function; if omitted or None, any hook function already installed is
        removed. The startup_hook function is called with no arguments just
        before readline prints the first prompt.

        '''
        self.mode.startup_hook = function

    def set_pre_input_hook(self, function=None):
        '''Set or remove the pre_input_hook function.

        If function is specified, it will be used as the new pre_input_hook
        function; if omitted or None, any hook function already installed is
        removed. The pre_input_hook function is called with no arguments
        after the first prompt has been printed and just before readline
        starts reading input characters.

        '''
        self.mode.pre_input_hook = function

#Functions that are not relevant for all Readlines but should at least have a NOP

    def _bell(self):
        pass

#
# Standard call, not available for all implementations
#
    
    def readline(self, prompt=''):
        raise NotImplementedError

#
# Callback interface
#
    def process_keyevent(self, keyinfo):
        return self.mode.process_keyevent(keyinfo)
        
    def readline_setup(self, prompt=""):
        return self.mode.readline_setup(prompt)

    def keyboard_poll(self):
        return self.mode._readline_from_keyboard_poll()

    def callback_handler_install(self, prompt, callback):
        '''bool readline_callback_handler_install ( string prompt, callback callback)
        Initializes the readline callback interface and terminal, prints the prompt and returns immediately
        '''
        self.callback = callback
        self.readline_setup(prompt)

    def callback_handler_remove(self):
        '''Removes a previously installed callback handler and restores terminal settings'''
        self.callback = None

    def callback_read_char(self):
        '''Reads a character and informs the readline callback interface when a line is received'''
        if self.keyboard_poll():
            line = self.get_line_buffer() + '\n'
            # however there is another newline added by
            # self.mode.readline_setup(prompt) which is called by callback_handler_install
            # this differs from GNU readline
            self.add_history(self.mode.l_buffer)
            # TADA:
            self.callback(line)

    def read_inputrc(self, #in 2.4 we cannot call expanduser with unicode string
                     inputrcpath=os.path.expanduser(ensure_str("~/pyreadlineconfig.ini"))):
        modes = dict([(x.mode,x) for x in self.editingmodes])
        mode = self.editingmodes[0].mode

        def setmode(name):
            self.mode = modes[name]

        def bind_key(key, name):
            import types
            if callable(name):
                modes[mode]._bind_key(key, types.MethodType(name, modes[mode]))
            elif hasattr(modes[mode], name):
                modes[mode]._bind_key(key, getattr(modes[mode], name))
            else:
                print("Trying to bind unknown command '%s' to key '%s'"%(name, key))

        def un_bind_key(key):
            keyinfo = make_KeyPress_from_keydescr(key).tuple()
            if keyinfo in modes[mode].key_dispatch:
                del modes[mode].key_dispatch[keyinfo]

        def bind_exit_key(key):
            modes[mode]._bind_exit_key(key)
            
        def un_bind_exit_key(key):
            keyinfo = make_KeyPress_from_keydescr(key).tuple()
            if keyinfo in modes[mode].exit_dispatch:
                del modes[mode].exit_dispatch[keyinfo]

        def setkill_ring_to_clipboard(killring):
            import pyreadline.lineeditor.lineobj 
            pyreadline.lineeditor.lineobj.kill_ring_to_clipboard = killring

        def sethistoryfilename(filename):
            self.mode._history.history_filename = os.path.expanduser(ensure_str(filename))

        def setbellstyle(mode):
            self.bell_style = mode

        def disable_readline(mode):
            self.disable_readline = mode

        def sethistorylength(length):
            self.mode._history.history_length = int(length)

        def allow_ctrl_c(mode):
            log("allow_ctrl_c:%s:%s"%(self.allow_ctrl_c, mode))
            self.allow_ctrl_c = mode
 
        def setbellstyle(mode):
            self.bell_style = mode
 
        def show_all_if_ambiguous(mode):
            self.mode.show_all_if_ambiguous = mode
        
        def ctrl_c_tap_time_interval(mode):
            self.ctrl_c_tap_time_interval = mode
        
        def mark_directories(mode):
            self.mode.mark_directories = mode
        
        def completer_delims(delims):
            self.mode.completer_delims = delims
        
        def complete_filesystem(delims):
            self.mode.complete_filesystem = delims.lower()

        def enable_ipython_paste_for_paths(boolean):
            self.mode.enable_ipython_paste_for_paths = boolean

        def debug_output(on, filename="pyreadline_debug_log.txt"): #Not implemented yet
            if on in ["on", "on_nologfile"]:
                self.debug=True

            if on == "on":
                logger.start_file_log(filename)
                logger.start_socket_log()
                logger.log("STARTING LOG")
            elif on == "on_nologfile":
                logger.start_socket_log()
                logger.log("STARTING LOG")
            else:
                logger.log("STOPING LOG")
                logger.stop_file_log()
                logger.stop_socket_log()
        
        _color_trtable={"black":0,      "darkred":4,  "darkgreen":2, 
                        "darkyellow":6, "darkblue":1, "darkmagenta":5,
                        "darkcyan":3,   "gray":7,     "red":4+8,
                        "green":2+8,    "yellow":6+8, "blue":1+8,
                        "magenta":5+8,  "cyan":3+8,   "white":7+8}
        
        def set_prompt_color(color):
            self.prompt_color = self._color_trtable.get(color.lower(),7)            
            
        def set_input_color(color):
            self.command_color=self._color_trtable.get(color.lower(),7)            

        loc = {"branch":release.branch,
               "version":release.version,
               "mode":mode,
               "modes":modes,
               "set_mode":setmode,
               "bind_key":bind_key,
               "disable_readline":disable_readline,
               "bind_exit_key":bind_exit_key,
               "un_bind_key":un_bind_key,
               "un_bind_exit_key":un_bind_exit_key,
               "bell_style":setbellstyle,
               "mark_directories":mark_directories,
               "show_all_if_ambiguous":show_all_if_ambiguous,
               "completer_delims":completer_delims,
               "complete_filesystem":complete_filesystem,
               "debug_output":debug_output,
               "history_filename":sethistoryfilename,
               "history_length":sethistorylength,
               "set_prompt_color":set_prompt_color,
               "set_input_color":set_input_color,
               "allow_ctrl_c":allow_ctrl_c,
               "ctrl_c_tap_time_interval":ctrl_c_tap_time_interval,
               "kill_ring_to_clipboard":setkill_ring_to_clipboard,
               "enable_ipython_paste_for_paths":enable_ipython_paste_for_paths,
              }
        if os.path.isfile(inputrcpath): 
            try:
                execfile(inputrcpath, loc, loc)
            except Exception as x:
                raise
                import traceback
                print("Error reading .pyinputrc", file=sys.stderr)
                filepath, lineno = traceback.extract_tb(sys.exc_traceback)[1][:2]
                print("Line: %s in file %s"%(lineno, filepath), file=sys.stderr)
                print(x, file=sys.stderr)
                raise ReadlineError("Error reading .pyinputrc")



class Readline(BaseReadline):
    """Baseclass for readline based on a console
    """
    def __init__(self):
        BaseReadline.__init__(self)
        self.console = console.Console()
        self.selection_color = self.console.saveattr<<4
        self.command_color = None
        self.prompt_color = None
        self.size = self.console.size()

        # variables you can control with parse_and_bind

#  To export as readline interface


##  Internal functions

    def _bell(self):
        '''ring the bell if requested.'''
        if self.bell_style == 'none':
            pass
        elif self.bell_style == 'visible':
            raise NotImplementedError("Bellstyle visible is not implemented yet.")
        elif self.bell_style == 'audible':
            self.console.bell()
        else:
            raise ReadlineError("Bellstyle %s unknown."%self.bell_style)

    def _clear_after(self):
        c = self.console
        x, y = c.pos()
        w, h = c.size()
        c.rectangle((x, y, w+1, y+1))
        c.rectangle((0, y+1, w, min(y+3,h)))

    def _set_cursor(self):
        c = self.console
        xc, yc = self.prompt_end_pos
        w, h = c.size()
        xc += self.mode.l_buffer.visible_line_width()
        while(xc >= w):
            xc -= w
            yc += 1
        c.pos(xc, yc)

    def _print_prompt(self):
        c = self.console
        x, y = c.pos()
        
        n = c.write_scrolling(self.prompt, self.prompt_color)
        self.prompt_begin_pos = (x, y - n)
        self.prompt_end_pos = c.pos()
        self.size = c.size()

    def _update_prompt_pos(self, n):
        if n != 0:
            bx, by = self.prompt_begin_pos
            ex, ey = self.prompt_end_pos
            self.prompt_begin_pos = (bx, by - n)
            self.prompt_end_pos = (ex, ey - n)

    def _update_line(self):
        c = self.console
        l_buffer = self.mode.l_buffer
        c.cursor(0)         #Hide cursor avoiding flicking
        c.pos(*self.prompt_begin_pos)
        self._print_prompt()
        ltext = l_buffer.quoted_text()
        if l_buffer.enable_selection and (l_buffer.selection_mark >= 0):
            start = len(l_buffer[:l_buffer.selection_mark].quoted_text())
            stop  = len(l_buffer[:l_buffer.point].quoted_text())
            if start > stop:
                stop,start = start,stop
            n = c.write_scrolling(ltext[:start], self.command_color)
            n = c.write_scrolling(ltext[start:stop], self.selection_color)
            n = c.write_scrolling(ltext[stop:], self.command_color)
        else:
            n = c.write_scrolling(ltext, self.command_color)

        x, y = c.pos()       #Preserve one line for Asian IME(Input Method Editor) statusbar
        w, h = c.size()
        if (y >= h - 1) or (n > 0):
            c.scroll_window(-1)
            c.scroll((0, 0, w, h), 0, -1)
            n += 1

        self._update_prompt_pos(n)
        if hasattr(c, "clear_to_end_of_window"): #Work around function for ironpython due 
            c.clear_to_end_of_window()          #to System.Console's lack of FillFunction
        else:
            self._clear_after()
        
        #Show cursor, set size vi mode changes size in insert/overwrite mode
        c.cursor(1, size=self.mode.cursor_size)  
        self._set_cursor()


    def callback_read_char(self):
        #Override base to get automatic newline
        '''Reads a character and informs the readline callback interface when a line is received'''
        if self.keyboard_poll():
            line = self.get_line_buffer() + '\n'
            self.console.write("\r\n")
            # however there is another newline added by
            # self.mode.readline_setup(prompt) which is called by callback_handler_install
            # this differs from GNU readline
            self.add_history(self.mode.l_buffer)
            # TADA:
            self.callback(line)


    def event_available(self):
        return self.console.peek() or (len(self.paste_line_buffer) > 0)

        
    def _readline_from_keyboard(self):
        while 1:
            if self._readline_from_keyboard_poll():
                break

    def _readline_from_keyboard_poll(self):
        pastebuffer = self.mode.paste_line_buffer
        if len(pastebuffer) > 0:
            #paste first line in multiline paste buffer
            self.l_buffer = lineobj.ReadLineTextBuffer(pastebuffer[0])
            self._update_line()
            self.mode.paste_line_buffer = pastebuffer[1:]
            return True

        c = self.console
        def nop(e):
            pass
        try:
            event = c.getkeypress()
        except KeyboardInterrupt:
            event = self.handle_ctrl_c()
        try:
            result = self.mode.process_keyevent(event.keyinfo)
        except EOFError:
            logger.stop_logging()
            raise
        self._update_line()
        return result

    def readline_setup(self, prompt=''):
        BaseReadline.readline_setup(self, prompt)
        self._print_prompt()
        self._update_line()

    def readline(self, prompt=''):
        self.readline_setup(prompt)
        self.ctrl_c_timeout = time.time()
        self._readline_from_keyboard()
        self.console.write('\r\n')
        log('returning(%s)' % self.get_line_buffer())
        return self.get_line_buffer() + '\n'

    def handle_ctrl_c(self):
        from pyreadline.keysyms.common import KeyPress
        from pyreadline.console.event import Event
        log("KBDIRQ")
        event = Event(0,0)
        event.char = "c"
        event.keyinfo = KeyPress("c", shift=False, control=True, 
                                 meta=False, keyname=None)
        if self.allow_ctrl_c:
            now = time.time()
            if (now - self.ctrl_c_timeout) < self.ctrl_c_tap_time_interval:
                log("Raise KeyboardInterrupt")
                raise KeyboardInterrupt
            else:
                self.ctrl_c_timeout = now
        else:
            raise KeyboardInterrupt
        return event




from __future__ import print_function, unicode_literals, absolute_import

FOREGROUND_BLUE           = 0x0001
FOREGROUND_GREEN          = 0x0002
FOREGROUND_RED            = 0x0004
FOREGROUND_INTENSITY      = 0x0008
BACKGROUND_BLUE           = 0x0010
BACKGROUND_GREEN          = 0x0020
BACKGROUND_RED            = 0x0040
BACKGROUND_INTENSITY      = 0x0080
COMMON_LVB_LEADING_BYTE   = 0x0100
COMMON_LVB_TRAILING_BYTE  = 0x0200
COMMON_LVB_GRID_HORIZONTAL= 0x0400
COMMON_LVB_GRID_LVERTICAL = 0x0800
COMMON_LVB_GRID_RVERTICAL = 0x1000
COMMON_LVB_REVERSE_VIDEO  = 0x2000
COMMON_LVB_UNDERSCORE     = 0x4000



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, absolute_import
"""Release data for the pyreadline project.

$Id$"""

#*****************************************************************************
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@kroywen.se>
#
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

# Name of the package for release purposes.  This is the name which labels
# the tarballs and RPMs made by distutils, so it's best to lowercase it.
name = 'pyreadline'

# For versions with substrings (like 0.6.16.svn), use an extra . to separate
# the new substring.  We have to avoid using either dashes or underscores,
# because bdist_rpm does not accept dashes (an RPM) convention, and
# bdist_deb does not accept underscores (a Debian convention).

branch = ''

version = '2.1'

description = "A python implmementation of GNU readline."

long_description = \
"""
The pyreadline package is a python implementation of GNU readline functionality
it is based on the ctypes based UNC readline package by Gary Bishop.
It is not complete. It has been tested for use with windows 2000 and windows xp.

* pyreadline 2.1 <2015-09-16>

  This is a bugfix release to make pyreadline work with python 3.5.

  Contributors to this release:

    - Jörgen Stenarson, maintainer
    - kivhift, improved error message for running on non windows platforms
    - zooba, made helpful suggestions to make it work for python 3.5

  Version 2.1 of pyreadline has been verfied for Python 2.7, and 3.4, 3.5.


Features:
 *  keyboard text selection and copy/paste
 *  Shift-arrowkeys for text selection
 *  Control-c can be used for copy activate with allow_ctrl_c(True) in config file
 *  Double tapping ctrl-c will raise a KeyboardInterrupt, use ctrl_c_tap_time_interval(x)
    where x is your preferred tap time window, default 0.3 s.
 *  paste pastes first line of content on clipboard.
 *  ipython_paste, pastes tab-separated data as list of lists or numpy array if all data is numeric
 *  paste_mulitline_code pastes multi line code, removing any empty lines.


 The latest development version is always available at the IPython github
 repository_.

.. _repository: https://github.com/pyreadline/pyreadline.git
 """

license = 'BSD'

authors = {'Jorgen' : ('Jorgen Stenarson','jorgen.stenarson@kroywen.se'),
           'Gary':    ('Gary Bishop', ''),
           'Jack':    ('Jack Trainor', ''),
           }

url = 'http://ipython.org/pyreadline.html'
download_url = 'https://pypi.python.org/pypi/pyreadline/'
platforms = ['Windows XP/2000/NT',
             'Windows 95/98/ME']

keywords = ['readline',
            'pyreadline']

classifiers = ['Development Status :: 5 - Production/Stable',
               'Environment :: Console',
               'Operating System :: Microsoft :: Windows',
               'License :: OSI Approved :: BSD License',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3.2',
               'Programming Language :: Python :: 3.3',
               'Programming Language :: Python :: 3.4',
               'Programming Language :: Python :: 3.5',
               ]



# -*- coding: utf-8 -*-
#*****************************************************************************
#       Copyright (C) 2003-2006 Gary Bishop.
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@bostream.nu>
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from __future__ import print_function, unicode_literals, absolute_import
import os, sys, time
import pyreadline.logger as logger
from   pyreadline.logger import log
from pyreadline.lineeditor.lineobj import Point
import pyreadline.lineeditor.lineobj as lineobj
import pyreadline.lineeditor.history as history
from . import basemode
from pyreadline.unicode_helper import ensure_unicode


def format(keyinfo):
    if len(keyinfo[-1]) != 1:
        k = keyinfo + (-1,)
    else:
        k = keyinfo + (ord(keyinfo[-1]),)
    return "(%s,%s,%s,%s,%x)"%k

in_ironpython = "IronPython" in sys.version


class IncrementalSearchPromptMode(object):
    def __init__(self, rlobj):
        pass

    def _process_incremental_search_keyevent(self, keyinfo):
        log("_process_incremental_search_keyevent")
        keytuple = keyinfo.tuple()
        #dispatch_func = self.key_dispatch.get(keytuple, default)
        revtuples = []
        fwdtuples = []
        for ktuple, func in self.key_dispatch.items():
            if func == self.reverse_search_history:
                revtuples.append(ktuple)
            elif func == self.forward_search_history:
                fwdtuples.append(ktuple)
        
        
        log("IncrementalSearchPromptMode %s %s"%(keyinfo, keytuple))
        if keyinfo.keyname == 'backspace':
            self.subsearch_query = self.subsearch_query[:-1]
            if len(self.subsearch_query) > 0:
                self.line = self.subsearch_fun(self.subsearch_query)
            else:
                self._bell()
                self.line = ""   # empty query means no search result
        elif keyinfo.keyname in ['return', 'escape']:
            self._bell()
            self.prompt = self.subsearch_oldprompt
            self.process_keyevent_queue = self.process_keyevent_queue[:-1]
            self._history.history_cursor = len(self._history.history)
            if keyinfo.keyname == 'escape':
                self.l_buffer.set_line(self.subsearch_old_line)
            return True
        elif keyinfo.keyname:
            pass
        elif keytuple in revtuples:
            self.subsearch_fun = self._history.reverse_search_history
            self.subsearch_prompt = "reverse-i-search%d`%s': "
            self.line = self.subsearch_fun(self.subsearch_query)
        elif keytuple in fwdtuples:
            self.subsearch_fun = self._history.forward_search_history
            self.subsearch_prompt = "forward-i-search%d`%s': "
            self.line = self.subsearch_fun(self.subsearch_query)
        elif keyinfo.control == False and keyinfo.meta == False:
            self.subsearch_query += keyinfo.char
            self.line = self.subsearch_fun(self.subsearch_query)
        else:
            pass
        self.prompt = self.subsearch_prompt%(self._history.history_cursor, self.subsearch_query)
        self.l_buffer.set_line(self.line)

    def _init_incremental_search(self, searchfun, init_event):
        """Initialize search prompt
        """
        log("init_incremental_search")
        self.subsearch_query = ''
        self.subsearch_fun = searchfun
        self.subsearch_old_line = self.l_buffer.get_line_text()

        queue = self.process_keyevent_queue
        queue.append(self._process_incremental_search_keyevent)

        self.subsearch_oldprompt = self.prompt

        if (self.previous_func != self.reverse_search_history and
            self.previous_func != self.forward_search_history):
            self.subsearch_query = self.l_buffer[0:Point].get_line_text()

        if self.subsearch_fun == self.reverse_search_history:
            self.subsearch_prompt = "reverse-i-search%d`%s': "
        else:
            self.subsearch_prompt = "forward-i-search%d`%s': "

        self.prompt = self.subsearch_prompt%(self._history.history_cursor, "")

        if self.subsearch_query:
            self.line = self._process_incremental_search_keyevent(init_event)
        else:
            self.line = ""


class SearchPromptMode(object):
    def __init__(self, rlobj):
        pass

    def _process_non_incremental_search_keyevent(self, keyinfo):
        keytuple = keyinfo.tuple()
        log("SearchPromptMode %s %s"%(keyinfo, keytuple))
        history = self._history

        if keyinfo.keyname == 'backspace':
            self.non_inc_query = self.non_inc_query[:-1]
        elif keyinfo.keyname in ['return', 'escape']:
            if self.non_inc_query:
                if self.non_inc_direction == -1:
                    res = history.reverse_search_history(self.non_inc_query)
                else:
                    res = history.forward_search_history(self.non_inc_query)

            self._bell()
            self.prompt = self.non_inc_oldprompt
            self.process_keyevent_queue = self.process_keyevent_queue[:-1]
            self._history.history_cursor = len(self._history.history)
            if keyinfo.keyname == 'escape':
                self.l_buffer = self.non_inc_oldline
            else:
                self.l_buffer.set_line(res)
            return False
        elif keyinfo.keyname:
            pass
        elif keyinfo.control == False and keyinfo.meta == False:
            self.non_inc_query += keyinfo.char
        else:
            pass
        self.prompt = self.non_inc_oldprompt + ":" + self.non_inc_query

    def _init_non_i_search(self, direction):
        self.non_inc_direction = direction
        self.non_inc_query = ""
        self.non_inc_oldprompt = self.prompt
        self.non_inc_oldline = self.l_buffer.copy()
        self.l_buffer.reset_line()
        self.prompt = self.non_inc_oldprompt + ":"
        queue = self.process_keyevent_queue
        queue.append(self._process_non_incremental_search_keyevent)

    def non_incremental_reverse_search_history(self, e):  # (M-p)
        '''Search backward starting at the current line and moving up
        through the history as necessary using a non-incremental search for
        a string supplied by the user.'''
        return self._init_non_i_search(-1)

    def non_incremental_forward_search_history(self, e):  # (M-n)
        '''Search forward starting at the current line and moving down
        through the the history as necessary using a non-incremental search
        for a string supplied by the user.'''
        return self._init_non_i_search(1)


class LeaveModeTryNext(Exception):
    pass


class DigitArgumentMode(object):
    def __init__(self, rlobj):
        pass

    def _process_digit_argument_keyevent(self, keyinfo):
        log("DigitArgumentMode.keyinfo %s"%keyinfo)
        keytuple = keyinfo.tuple()
        log("DigitArgumentMode.keytuple %s %s"%(keyinfo, keytuple))
        if keyinfo.keyname in ['return']:
            self.prompt = self._digit_argument_oldprompt
            self.process_keyevent_queue = self.process_keyevent_queue[:-1]
            return True
        elif keyinfo.keyname:
            pass
        elif (keyinfo.char in "0123456789" and
              keyinfo.control == False and
              keyinfo.meta == False):
            log("arg %s %s"%(self.argument, keyinfo.char))
            self.argument = self.argument * 10 + int(keyinfo.char)
        else:
            self.prompt = self._digit_argument_oldprompt
            raise LeaveModeTryNext
        self.prompt = "(arg: %s) "%self.argument

    def _init_digit_argument(self, keyinfo):
        """Initialize search prompt
        """
        c = self.console
        line = self.l_buffer.get_line_text()
        self._digit_argument_oldprompt = self.prompt
        queue = self.process_keyevent_queue
        queue = self.process_keyevent_queue
        queue.append(self._process_digit_argument_keyevent)

        if keyinfo.char == "-":
            self.argument = -1
        elif keyinfo.char in "0123456789":
            self.argument = int(keyinfo.char)
        log("<%s> %s"%(self.argument, type(self.argument)))
        self.prompt = "(arg: %s) "%self.argument
        log("arg-init %s %s"%(self.argument, keyinfo.char))


class EmacsMode(DigitArgumentMode, IncrementalSearchPromptMode,
                SearchPromptMode, basemode.BaseMode):
    mode = "emacs"

    def __init__(self, rlobj):
        basemode.BaseMode.__init__(self, rlobj)
        IncrementalSearchPromptMode.__init__(self, rlobj)
        SearchPromptMode.__init__(self, rlobj)
        DigitArgumentMode.__init__(self, rlobj)
        self._keylog = (lambda x, y: None)
        self.previous_func = None
        self.prompt = ">>> "
        self._insert_verbatim = False
        self.next_meta = False  # True to force meta on next character

        self.process_keyevent_queue = [self._process_keyevent]

    def __repr__(self):
        return "<EmacsMode>"

    def add_key_logger(self, logfun):
        """logfun should be function that takes disp_fun and line_"""\
        """buffer object """
        self._keylog = logfun

    def process_keyevent(self, keyinfo):
        try:
            r = self.process_keyevent_queue[-1](keyinfo)
        except LeaveModeTryNext:
            self.process_keyevent_queue = self.process_keyevent_queue[:-1]
            r = self.process_keyevent(keyinfo)
        if r:
            self.add_history(self.l_buffer.copy())
            return True
        return False

    def _process_keyevent(self, keyinfo):
        """return True when line is final
        """
        #Process exit keys. Only exit on empty line
        log("_process_keyevent <%s>"%keyinfo)

        def nop(e):
            pass
        if self.next_meta:
            self.next_meta = False
            keyinfo.meta = True
        keytuple = keyinfo.tuple()

        if self._insert_verbatim:
            self.insert_text(keyinfo)
            self._insert_verbatim = False
            self.argument = 0
            return False

        if keytuple in self.exit_dispatch:
            pars = (self.l_buffer, lineobj.EndOfLine(self.l_buffer))
            log("exit_dispatch:<%s, %s>"%pars)
            if lineobj.EndOfLine(self.l_buffer) == 0:
                raise EOFError
        if keyinfo.keyname or keyinfo.control or keyinfo.meta:
            default = nop
        else:
            default = self.self_insert
        dispatch_func = self.key_dispatch.get(keytuple, default)

        log("readline from keyboard:<%s,%s>"%(keytuple, dispatch_func))

        r = None
        if dispatch_func:
            r = dispatch_func(keyinfo)
            self._keylog(dispatch_func, self.l_buffer)
            self.l_buffer.push_undo()

        self.previous_func = dispatch_func
        return r

#########  History commands
    def previous_history(self, e):  # (C-p)
        '''Move back through the history list, fetching the previous
        command. '''
        self._history.previous_history(self.l_buffer)
        self.l_buffer.point = lineobj.EndOfLine
        self.finalize()

    def next_history(self, e):  # (C-n)
        '''Move forward through the history list, fetching the next
        command. '''
        self._history.next_history(self.l_buffer)
        self.finalize()

    def beginning_of_history(self, e):  # (M-<)
        '''Move to the first line in the history.'''
        self._history.beginning_of_history()
        self.finalize()

    def end_of_history(self, e):  # (M->)
        '''Move to the end of the input history, i.e., the line currently
        being entered.'''
        self._history.end_of_history(self.l_buffer)
        self.finalize()

    def reverse_search_history(self, e):  # (C-r)
        '''Search backward starting at the current line and moving up
        through the history as necessary. This is an incremental search.'''
        log("rev_search_history")
        self._init_incremental_search(self._history.reverse_search_history, e)
        self.finalize()

    def forward_search_history(self, e):  # (C-s)
        '''Search forward starting at the current line and moving down
        through the the history as necessary. This is an incremental
        search.'''
        log("fwd_search_history")
        self._init_incremental_search(self._history.forward_search_history, e)
        self.finalize()

    def history_search_forward(self, e):  # ()
        '''Search forward through the history for the string of characters
        between the start of the current line and the point. This is a
        non-incremental search. By default, this command is unbound.'''
        if (self.previous_func and
            hasattr(self._history, self.previous_func.__name__)):
            self._history.lastcommand = getattr(self._history,
                                                self.previous_func.__name__)
        else:
            self._history.lastcommand = None
        q = self._history.history_search_forward(self.l_buffer)
        self.l_buffer = q
        self.l_buffer.point = q.point
        self.finalize()

    def history_search_backward(self, e):  # ()
        '''Search backward through the history for the string of characters
        between the start of the current line and the point. This is a
        non-incremental search. By default, this command is unbound.'''
        if (self.previous_func and
            hasattr(self._history, self.previous_func.__name__)):
            self._history.lastcommand = getattr(self._history,
                                                self.previous_func.__name__)
        else:
            self._history.lastcommand = None
        q = self._history.history_search_backward(self.l_buffer)
        self.l_buffer = q
        self.l_buffer.point = q.point
        self.finalize()

    def yank_nth_arg(self, e):  # (M-C-y)
        '''Insert the first argument to the previous command (usually the
        second word on the previous line) at point. With an argument n,
        insert the nth word from the previous command (the words in the
        previous command begin with word 0). A negative argument inserts the
        nth word from the end of the previous command.'''
        self.finalize()

    def yank_last_arg(self, e):  # (M-. or M-_)
        '''Insert last argument to the previous command (the last word of
        the previous history entry). With an argument, behave exactly like
        yank-nth-arg. Successive calls to yank-last-arg move back through
        the history list, inserting the last argument of each line in turn.'''
        self.finalize()

    def forward_backward_delete_char(self, e):  # ()
        '''Delete the character under the cursor, unless the cursor is at
        the end of the line, in which case the character behind the cursor
        is deleted. By default, this is not bound to a key.'''
        self.finalize()

    def quoted_insert(self, e):  # (C-q or C-v)
        '''Add the next character typed to the line verbatim. This is how to
        insert key sequences like C-q, for example.'''
        self._insert_verbatim = True
        self.finalize()

    def tab_insert(self, e):  # (M-TAB)
        '''Insert a tab character. '''
        cursor = min(self.l_buffer.point, len(self.l_buffer.line_buffer))
        ws = ' ' * (self.tabstop - (cursor % self.tabstop))
        self.insert_text(ws)
        self.finalize()

    def transpose_chars(self, e):  # (C-t)
        '''Drag the character before the cursor forward over the character
        at the cursor, moving the cursor forward as well. If the insertion
        point is at the end of the line, then this transposes the last two
        characters of the line. Negative arguments have no effect.'''
        self.l_buffer.transpose_chars()
        self.finalize()

    def transpose_words(self, e):  # (M-t)
        '''Drag the word before point past the word after point, moving
        point past that word as well. If the insertion point is at the end
        of the line, this transposes the last two words on the line.'''
        self.l_buffer.transpose_words()
        self.finalize()

    def overwrite_mode(self, e):  # ()
        '''Toggle overwrite mode. With an explicit positive numeric
        argument, switches to overwrite mode. With an explicit non-positive
        numeric argument, switches to insert mode. This command affects only
        emacs mode; vi mode does overwrite differently. Each call to
        readline() starts in insert mode. In overwrite mode, characters
        bound to self-insert replace the text at point rather than pushing
        the text to the right. Characters bound to backward-delete-char
        replace the character before point with a space.'''
        self.finalize()

    def kill_line(self, e):  # (C-k)
        '''Kill the text from point to the end of the line. '''
        self.l_buffer.kill_line()
        self.finalize()

    def backward_kill_line(self, e):  # (C-x Rubout)
        '''Kill backward to the beginning of the line. '''
        self.l_buffer.backward_kill_line()
        self.finalize()

    def unix_line_discard(self, e):  # (C-u)
        '''Kill backward from the cursor to the beginning of the current
        line. '''
        # how is this different from backward_kill_line?
        self.l_buffer.unix_line_discard()
        self.finalize()

    def kill_whole_line(self, e):  # ()
        '''Kill all characters on the current line, no matter where point
        is. By default, this is unbound.'''
        self.l_buffer.kill_whole_line()
        self.finalize()

    def kill_word(self, e):  # (M-d)
        '''Kill from point to the end of the current word, or if between
        words, to the end of the next word. Word boundaries are the same as
        forward-word.'''
        self.l_buffer.kill_word()
        self.finalize()

    forward_kill_word = kill_word

    def backward_kill_word(self, e):  # (M-DEL)
        '''Kill the word behind point. Word boundaries are the same as
        backward-word. '''
        self.l_buffer.backward_kill_word()
        self.finalize()

    def unix_word_rubout(self, e):  # (C-w)
        '''Kill the word behind point, using white space as a word
        boundary. The killed text is saved on the kill-ring.'''
        self.l_buffer.unix_word_rubout()
        self.finalize()

    def kill_region(self, e):  # ()
        '''Kill the text in the current region. By default, this command is
        unbound. '''
        self.finalize()

    def copy_region_as_kill(self, e):  # ()
        '''Copy the text in the region to the kill buffer, so it can be
        yanked right away. By default, this command is unbound.'''
        self.finalize()

    def copy_backward_word(self, e):  # ()
        '''Copy the word before point to the kill buffer. The word
        boundaries are the same as backward-word. By default, this command
        is unbound.'''
        self.finalize()

    def copy_forward_word(self, e):  # ()
        '''Copy the word following point to the kill buffer. The word
        boundaries are the same as forward-word. By default, this command is
        unbound.'''
        self.finalize()

    def yank(self, e):  # (C-y)
        '''Yank the top of the kill ring into the buffer at point. '''
        self.l_buffer.yank()
        self.finalize()

    def yank_pop(self, e):  # (M-y)
        '''Rotate the kill-ring, and yank the new top. You can only do this
        if the prior command is yank or yank-pop.'''
        self.l_buffer.yank_pop()
        self.finalize()

    def delete_char_or_list(self, e):  # ()
        '''Deletes the character under the cursor if not at the beginning or
        end of the line (like delete-char). If at the end of the line,
        behaves identically to possible-completions. This command is unbound
        by default.'''
        self.finalize()

    def start_kbd_macro(self, e):  # (C-x ()
        '''Begin saving the characters typed into the current keyboard
        macro. '''
        self.finalize()

    def end_kbd_macro(self, e):  # (C-x ))
        '''Stop saving the characters typed into the current keyboard macro
        and save the definition.'''
        self.finalize()

    def call_last_kbd_macro(self, e):  # (C-x e)
        '''Re-execute the last keyboard macro defined, by making the
        characters in the macro appear as if typed at the keyboard.'''
        self.finalize()

    def re_read_init_file(self, e):  # (C-x C-r)
        '''Read in the contents of the inputrc file, and incorporate any
        bindings or variable assignments found there.'''
        self.finalize()

    def abort(self, e):  # (C-g)
        '''Abort the current editing command and ring the terminals bell
             (subject to the setting of bell-style).'''
        self._bell()
        self.finalize()

    def do_uppercase_version(self, e):  # (M-a, M-b, M-x, ...)
        '''If the metafied character x is lowercase, run the command that is
        bound to the corresponding uppercase character.'''
        self.finalize()

    def prefix_meta(self, e):  # (ESC)
        '''Metafy the next character typed. This is for keyboards without a
        meta key. Typing ESC f is equivalent to typing M-f. '''
        self.next_meta = True
        self.finalize()

    def undo(self, e):  # (C-_ or C-x C-u)
        '''Incremental undo, separately remembered for each line.'''
        self.l_buffer.pop_undo()
        self.finalize()

    def revert_line(self, e):  # (M-r)
        '''Undo all changes made to this line. This is like executing the
        undo command enough times to get back to the beginning.'''
        self.finalize()

    def tilde_expand(self, e):  # (M-~)
        '''Perform tilde expansion on the current word.'''
        self.finalize()

    def set_mark(self, e):  # (C-@)
        '''Set the mark to the point. If a numeric argument is supplied, the
        mark is set to that position.'''
        self.l_buffer.set_mark()
        self.finalize()

    def exchange_point_and_mark(self, e):  # (C-x C-x)
        '''Swap the point with the mark. The current cursor position is set
        to the saved position, and the old cursor position is saved as the
        mark.'''
        self.finalize()

    def character_search(self, e):  # (C-])
        '''A character is read and point is moved to the next occurrence of
        that character. A negative count searches for previous occurrences.'''
        self.finalize()

    def character_search_backward(self, e):  # (M-C-])
        '''A character is read and point is moved to the previous occurrence
        of that character. A negative count searches for subsequent
        occurrences.'''
        self.finalize()

    def insert_comment(self, e):  # (M-#)
        '''Without a numeric argument, the value of the comment-begin
        variable is inserted at the beginning of the current line. If a
        numeric argument is supplied, this command acts as a toggle: if the
        characters at the beginning of the line do not match the value of
        comment-begin, the value is inserted, otherwise the characters in
        comment-begin are deleted from the beginning of the line. In either
        case, the line is accepted as if a newline had been typed.'''
        self.finalize()

    def dump_variables(self, e):  # ()
        '''Print all of the settable variables and their values to the
        Readline output stream. If a numeric argument is supplied, the
        output is formatted in such a way that it can be made part of an
        inputrc file. This command is unbound by default.'''
        self.finalize()

    def dump_macros(self, e):  # ()
        '''Print all of the Readline key sequences bound to macros and the
        strings they output. If a numeric argument is supplied, the output
        is formatted in such a way that it can be made part of an inputrc
        file. This command is unbound by default.'''
        self.finalize()

    def digit_argument(self, e):  # (M-0, M-1, ... M--)
        '''Add this digit to the argument already accumulating, or start a
        new argument. M-- starts a negative argument.'''
        self._init_digit_argument(e)
        #Should not finalize

    def universal_argument(self, e):  # ()
        '''This is another way to specify an argument. If this command is
        followed by one or more digits, optionally with a leading minus
        sign, those digits define the argument. If the command is followed
        by digits, executing universal-argument again ends the numeric
        argument, but is otherwise ignored. As a special case, if this
        command is immediately followed by a character that is neither a
        digit or minus sign, the argument count for the next command is
        multiplied by four. The argument count is initially one, so
        executing this function the first time makes the argument count
        four, a second time makes the argument count sixteen, and so on. By
        default, this is not bound to a key.'''
        #Should not finalize

    #Create key bindings:
    def init_editing_mode(self, e):  # (C-e)
        '''When in vi command mode, this causes a switch to emacs editing
        mode.'''
        self._bind_exit_key('Control-d')
        self._bind_exit_key('Control-z')

        # I often accidentally hold the shift or control while typing space
        self._bind_key('space',       self.self_insert)
        self._bind_key('Shift-space',       self.self_insert)
        self._bind_key('Control-space',     self.self_insert)
        self._bind_key('Return',            self.accept_line)
        self._bind_key('Left',              self.backward_char)
        self._bind_key('Control-b',         self.backward_char)
        self._bind_key('Right',             self.forward_char)
        self._bind_key('Control-f',         self.forward_char)
        self._bind_key('Control-h',         self.backward_delete_char)
        self._bind_key('BackSpace',         self.backward_delete_char)
        self._bind_key('Control-BackSpace', self.backward_delete_word)

        self._bind_key('Home',              self.beginning_of_line)
        self._bind_key('End',               self.end_of_line)
        self._bind_key('Delete',            self.delete_char)
        self._bind_key('Control-d',         self.delete_char)
        self._bind_key('Clear',             self.clear_screen)
        self._bind_key('Alt-f',             self.forward_word)
        self._bind_key('Alt-b',             self.backward_word)
        self._bind_key('Control-l',         self.clear_screen)
        self._bind_key('Control-p',         self.previous_history)
        self._bind_key('Up',                self.history_search_backward)
        self._bind_key('Control-n',         self.next_history)
        self._bind_key('Down',              self.history_search_forward)
        self._bind_key('Control-a',         self.beginning_of_line)
        self._bind_key('Control-e',         self.end_of_line)
        self._bind_key('Alt-<',             self.beginning_of_history)
        self._bind_key('Alt->',             self.end_of_history)
        self._bind_key('Control-r',         self.reverse_search_history)
        self._bind_key('Control-s',         self.forward_search_history)
        self._bind_key('Control-Shift-r',         self.forward_search_history)
        self._bind_key('Alt-p',
                       self.non_incremental_reverse_search_history)
        self._bind_key('Alt-n',
                       self.non_incremental_forward_search_history)
        self._bind_key('Control-z',         self.undo)
        self._bind_key('Control-_',         self.undo)
        self._bind_key('Escape',            self.kill_whole_line)
        self._bind_key('Meta-d',            self.kill_word)
        self._bind_key('Control-Delete',       self.forward_delete_word)
        self._bind_key('Control-w',         self.unix_word_rubout)
        #self._bind_key('Control-Shift-v',   self.quoted_insert)
        self._bind_key('Control-v',         self.paste)
        self._bind_key('Alt-v',             self.ipython_paste)
        self._bind_key('Control-y',         self.yank)
        self._bind_key('Control-k',         self.kill_line)
        self._bind_key('Control-m',         self.set_mark)
        self._bind_key('Control-q',         self.copy_region_to_clipboard)
#        self._bind_key('Control-shift-k',  self.kill_whole_line)
        self._bind_key('Control-Shift-v',   self.paste_mulitline_code)
        self._bind_key("Control-Right",     self.forward_word_end)
        self._bind_key("Control-Left",      self.backward_word)
        self._bind_key("Shift-Right",
                       self.forward_char_extend_selection)
        self._bind_key("Shift-Left",
                       self.backward_char_extend_selection)
        self._bind_key("Shift-Control-Right",
                       self.forward_word_end_extend_selection)
        self._bind_key("Shift-Control-Left",
                       self.backward_word_extend_selection)
        self._bind_key("Shift-Home",
                       self.beginning_of_line_extend_selection)
        self._bind_key("Shift-End",
                       self.end_of_line_extend_selection)
        self._bind_key("numpad0",           self.self_insert)
        self._bind_key("numpad1",           self.self_insert)
        self._bind_key("numpad2",           self.self_insert)
        self._bind_key("numpad3",           self.self_insert)
        self._bind_key("numpad4",           self.self_insert)
        self._bind_key("numpad5",           self.self_insert)
        self._bind_key("numpad6",           self.self_insert)
        self._bind_key("numpad7",           self.self_insert)
        self._bind_key("numpad8",           self.self_insert)
        self._bind_key("numpad9",           self.self_insert)
        self._bind_key("add",               self.self_insert)
        self._bind_key("subtract",          self.self_insert)
        self._bind_key("multiply",          self.self_insert)
        self._bind_key("divide",            self.self_insert)
        self._bind_key("vk_decimal",        self.self_insert)
        log("RUNNING INIT EMACS")
        for i in range(0, 10):
            self._bind_key("alt-%d"%i,      self.digit_argument)
        self._bind_key("alt--",             self.digit_argument)


# make it case insensitive
def commonprefix(m):
    "Given a list of pathnames, returns the longest common leading component"
    if not m:
        return ''
    prefix = m[0]
    for item in m:
        for i in range(len(prefix)):
            if prefix[:i + 1].lower() != item[:i + 1].lower():
                prefix = prefix[:i]
                if i == 0:
                    return ''
                break
    return prefix



# Copyright (C) 2006  Michael Graz. <mgraz@plan10.com>
from __future__ import print_function, unicode_literals, absolute_import

import sys, unittest
sys.path.append ('../..')
#from pyreadline.modes.vi import *
#from pyreadline import keysyms
from pyreadline.lineeditor import lineobj

#----------------------------------------------------------------------


#----------------------------------------------------------------------

class Test_copy (unittest.TestCase):
    def test_copy1 (self):
        l=lineobj.ReadLineTextBuffer("first second")
        q=l.copy()
        self.assertEqual(q.get_line_text(),l.get_line_text())
        self.assertEqual(q.point,l.point)
        self.assertEqual(q.mark,l.mark)
        
    def test_copy2 (self):
        l=lineobj.ReadLineTextBuffer("first second",point=5)
        q=l.copy()
        self.assertEqual(q.get_line_text(),l.get_line_text())
        self.assertEqual(q.point,l.point)
        self.assertEqual(q.mark,l.mark)
        

class Test_linepos (unittest.TestCase):
    t="test text"
    def test_NextChar (self):
        t=self.t
        l=lineobj.ReadLineTextBuffer(t)
        for i in range(len(t)):
            self.assertEqual(i,l.point)
            l.point=lineobj.NextChar
        #advance past end of buffer            
        l.point=lineobj.NextChar
        self.assertEqual(len(t),l.point)

    def test_PrevChar (self):
        t=self.t
        l=lineobj.ReadLineTextBuffer(t,point=len(t))
        for i in range(len(t)):
            self.assertEqual(len(t)-i,l.point)
            l.point=lineobj.PrevChar
        #advance past beginning of buffer            
        l.point=lineobj.PrevChar
        self.assertEqual(0,l.point)

    def test_EndOfLine (self):
        t=self.t
        l=lineobj.ReadLineTextBuffer(t,point=len(t))
        for i in range(len(t)):
            l.point=i
            l.point=lineobj.EndOfLine
            self.assertEqual(len(t),l.point)
                        
    def test_StartOfLine (self):
        t=self.t
        l=lineobj.ReadLineTextBuffer(t,point=len(t))
        for i in range(len(t)):
            l.point=i
            l.point=lineobj.StartOfLine
            self.assertEqual(0,l.point)
                        

class Tests_linepos2(Test_linepos):
    t="kajkj"
            
class Tests_linepos3(Test_linepos):
    t=""


class Test_movement (unittest.TestCase):
    def test_NextChar (self):
        cmd=lineobj.NextChar
        tests=[
        # "First"
         (cmd,
          "First",
          "#     ",
          " #    "),
         (cmd,
          "First",
          "    # ",
          "     #"),
         (cmd,
          "First",
          "     #",
          "     #"),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)
        
    def test_PrevChar (self):
        cmd=lineobj.PrevChar
        tests=[
        # "First"
         (cmd,
          "First",
          "     #",
          "    # "),
         (cmd,
          "First",
          " #   ",
          "#    "),
         (cmd,
          "First",
          "#     ",
          "#     "),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)
        
        
        
    def test_PrevWordStart (self):
        cmd=lineobj.PrevWordStart
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "                  #",
          "             #     "),
         (cmd,
          "First Second Third",
          "             #     ",
          "      #            "),
         (cmd,
          "First Second Third",
          "     #             ",
          "#                  "),
         (cmd,
          "First Second Third",
          "#                  ",
          "#                  "),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)
        
    def test_NextWordStart (self):
        cmd=lineobj.NextWordStart
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "#                 ",
          "      #           "),
         (cmd,
          "First Second Third",
          "    #             ",
          "      #           "),
         (cmd,
          "First Second Third",
          "      #            ",
          "             #     "),
         (cmd,
          "First Second Third",
          "              #    ",
          "                  #"),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)
        
    def test_NextWordEnd (self):
        cmd=lineobj.NextWordEnd
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "#                 ",
          "     #            "),
         (cmd,
          "First Second Third",
          "    #             ",
          "     #            "),
         (cmd,
          "First Second Third",
          "      #            ",
          "            #      "),
         (cmd,
          "First Second Third",
          "              #    ",
          "                  #"),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)
        
    def test_PrevWordEnd (self):
        cmd=lineobj.PrevWordEnd
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "                  #",
          "            #      "),
         (cmd,
          "First Second Third",
          "            #      ",
          "     #             "),
         (cmd,
          "First Second Third",
          "     #             ",
          "#                  "),
         (cmd,
          "First Second Third",
          "#                  ",
          "#                  "),          
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)
        
    def test_WordEnd_1 (self):
        cmd=lineobj.WordEnd
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "#                  ",
          "     #             "),
         (cmd,
          "First Second Third",
          " #                 ",
          "     #             "),
         (cmd,
          "First Second Third",
          "             #     ",
          "                  #"),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)

    def test_WordEnd_2 (self):
        cmd=lineobj.WordEnd
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "     #             "),
         (cmd,
          "First Second Third",
          "            #      "),
         (cmd,
          "First Second Third",
          "                  #"),
          ]

        for cmd,text,init_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            self.assertRaises(lineobj.NotAWordError,cmd,l)
        
        
    def test_WordStart_1 (self):
        cmd=lineobj.WordStart
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "#                  ",
          "#                  "),
         (cmd,
          "First Second Third",
          " #                 ",
          "#                  "),
         (cmd,
          "First Second Third",
          "               #   ",
          "             #     "),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)
        
    def test_WordStart_2 (self):
        cmd=lineobj.WordStart
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "     #             "),
         (cmd,
          "First Second Third",
          "            #      "),
         (cmd,
          "First Second Third",
          "                  #"),
          ]

        for cmd,text,init_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            self.assertRaises(lineobj.NotAWordError,cmd,l)
        

    def test_StartOfLine (self):
        cmd=lineobj.StartOfLine
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "#                 ",
          "#                 "),
         (cmd,
          "First Second Third",
          "         #         ",
          "#                  "),
         (cmd,
          "First Second Third",
          "                  #",
          "#                  "),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)
        
    def test_EndOfLine (self):
        cmd=lineobj.EndOfLine
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          "#                 ",
          "                  #"),
         (cmd,
          "First Second Third",
          "         #         ",
          "                  #"),
         (cmd,
          "First Second Third",
          "                  #",
          "                  #"),
          ]
        for cmd,text,init_point,expected_point in tests:
            l=lineobj.ReadLineTextBuffer(text,get_point_pos(init_point))
            l.point=cmd
            self.assertEqual(get_point_pos(expected_point),l.point)

    def test_Point(self):
        cmd=lineobj.Point
        tests=[
        # "First Second Third"
         (cmd,
          "First Second Third",
          0),
         (cmd,
          "First Second Third",
          12),
         (cmd,
          "First Second Third",
          18),
          ]
        for cmd,text,p in tests:
            l=lineobj.ReadLineTextBuffer(text,p)
            self.assertEqual(p,cmd(l))


#----------------------------------------------------------------------
# utility functions

def get_point_pos(pstr):
    return pstr.index("#")

def get_mark_pos(mstr):
    try:
        return mstr.index("#")
    except ValueError:
        return -1
#----------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()

    l=lineobj.ReadLineTextBuffer("First Second Third")



# -*- coding: utf-8 -*-
#*****************************************************************************
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@bostream.nu>
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from __future__ import print_function, unicode_literals, absolute_import
import clr
clr.AddReferenceByPartialName("System.Windows.Forms")
import System.Windows.Forms.Clipboard as cb

def GetClipboardText():
    text = ""
    if cb.ContainsText():
        text = cb.GetText()

    return text

def SetClipboardText(text):
    cb.SetText(text)    

if __name__ == '__main__':
    txt = GetClipboardText()                            # display last text clipped
    print(txt)
     
     
     
     


from __future__ import print_function, unicode_literals, absolute_import
import sys

if sys.version_info[0] >= 3:
    import collections
    PY3 = True
    def callable(x):
        return isinstance(x, collections.Callable)
    
    def execfile(fname, glob, loc=None):
        loc = loc if (loc is not None) else glob
        with open(fname) as fil:
            txt = fil.read()
        exec(compile(txt, fname, 'exec'), glob, loc)

    unicode = str
    bytes = bytes
    from io import StringIO
else:
    PY3 = False
    callable = callable
    execfile = execfile
    bytes = str
    unicode = unicode
    
    from StringIO import StringIO



# -*- coding: utf-8 -*-
#*****************************************************************************
#       Copyright (C) 2003-2006 Gary Bishop.
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@bostream.nu>
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from __future__ import print_function, unicode_literals, absolute_import
import System
from .common import validkey, KeyPress, make_KeyPress_from_keydescr

c32 = System.ConsoleKey
Shift = System.ConsoleModifiers.Shift
Control = System.ConsoleModifiers.Control
Alt = System.ConsoleModifiers.Alt
# table for translating virtual keys to X windows key symbols
code2sym_map = {#c32.CANCEL:    'Cancel',
                c32.Backspace:  'BackSpace',
                c32.Tab:        'Tab',
                c32.Clear:      'Clear',
                c32.Enter:      'Return',
#                c32.Shift:     'Shift_L',
#                c32.Control:   'Control_L',
#                c32.Menu:      'Alt_L',
                c32.Pause:      'Pause',
#                c32.Capital:   'Caps_Lock',
                c32.Escape:     'Escape',
#                c32.Space:     'space',
                c32.PageUp:     'Prior',
                c32.PageDown:   'Next',
                c32.End:        'End',
                c32.Home:       'Home',
                c32.LeftArrow:  'Left',
                c32.UpArrow:    'Up',
                c32.RightArrow: 'Right',
                c32.DownArrow:  'Down',
                c32.Select:     'Select',
                c32.Print:      'Print',
                c32.Execute:    'Execute',
#                c32.Snapshot:  'Snapshot',
                c32.Insert:     'Insert',
                c32.Delete:     'Delete',
                c32.Help:       'Help',
                c32.F1:         'F1',
                c32.F2:         'F2',
                c32.F3:         'F3',
                c32.F4:         'F4',
                c32.F5:         'F5',
                c32.F6:         'F6',
                c32.F7:         'F7',
                c32.F8:         'F8',
                c32.F9:         'F9',
                c32.F10:        'F10',
                c32.F11:        'F11',
                c32.F12:        'F12',
                c32.F13:        'F13',
                c32.F14:        'F14',
                c32.F15:        'F15',
                c32.F16:        'F16',
                c32.F17:        'F17',
                c32.F18:        'F18',
                c32.F19:        'F19',
                c32.F20:        'F20',
                c32.F21:        'F21',
                c32.F22:        'F22',
                c32.F23:        'F23',
                c32.F24:        'F24',
#                c32.Numlock:    'Num_Lock,',
#                c32.Scroll:     'Scroll_Lock',
#                c32.Apps:       'VK_APPS',
#                c32.ProcesskeY: 'VK_PROCESSKEY',
#                c32.Attn:       'VK_ATTN',
#                c32.Crsel:      'VK_CRSEL',
#                c32.Exsel:      'VK_EXSEL',
#                c32.Ereof:      'VK_EREOF',
#                c32.Play:       'VK_PLAY',
#                c32.Zoom:       'VK_ZOOM',
#                c32.Noname:     'VK_NONAME',
#                c32.Pa1:        'VK_PA1',
                c32.OemClear:  'VK_OEM_CLEAR',
                c32.NumPad0:    'NUMPAD0',
                c32.NumPad1:    'NUMPAD1',
                c32.NumPad2:    'NUMPAD2',
                c32.NumPad3:    'NUMPAD3',
                c32.NumPad4:    'NUMPAD4',
                c32.NumPad5:    'NUMPAD5',
                c32.NumPad6:    'NUMPAD6',
                c32.NumPad7:    'NUMPAD7',
                c32.NumPad8:    'NUMPAD8',
                c32.NumPad9:    'NUMPAD9',
                c32.Divide:     'Divide',
                c32.Multiply:   'Multiply',
                c32.Add:        'Add',
                c32.Subtract:   'Subtract',
                c32.Decimal:    'VK_DECIMAL'
               }

# function to handle the mapping
def make_keysym(keycode):
    try:
        sym = code2sym_map[keycode]
    except KeyError:
        sym = ''
    return sym

sym2code_map = {}
for code,sym in code2sym_map.items():
    sym2code_map[sym.lower()] = code

def key_text_to_keyinfo(keytext):
    '''Convert a GNU readline style textual description of a key to keycode with modifiers'''
    if keytext.startswith('"'): # "
        return keyseq_to_keyinfo(keytext[1:-1])
    else:
        return keyname_to_keyinfo(keytext)


def char_to_keyinfo(char, control=False, meta=False, shift=False):
    vk = (ord(char))
    if vk & 0xffff == 0xffff:
        print('VkKeyScan("%s") = %x' % (char, vk))
        raise ValueError('bad key')
    if vk & 0x100:
        shift = True
    if vk & 0x200:
        control = True
    if vk & 0x400:
        meta = True
    return (control, meta, shift, vk & 0xff)

def keyname_to_keyinfo(keyname):
    control = False
    meta = False
    shift = False

    while 1:
        lkeyname = keyname.lower()
        if lkeyname.startswith('control-'):
            control = True
            keyname = keyname[8:]
        elif lkeyname.startswith('ctrl-'):
            control = True
            keyname = keyname[5:]
        elif lkeyname.startswith('meta-'):
            meta = True
            keyname = keyname[5:]
        elif lkeyname.startswith('alt-'):
            meta = True
            keyname = keyname[4:]
        elif lkeyname.startswith('shift-'):
            shift = True
            keyname = keyname[6:]
        else:
            if len(keyname) > 1:
                return (control, meta, shift, sym2code_map.get(keyname.lower()," "))
            else:
                return char_to_keyinfo(keyname, control, meta, shift)

def keyseq_to_keyinfo(keyseq):
    res = []
    control = False
    meta = False
    shift = False

    while 1:
        if keyseq.startswith('\\C-'):
            control = True
            keyseq = keyseq[3:]
        elif keyseq.startswith('\\M-'):
            meta = True
            keyseq = keyseq[3:]
        elif keyseq.startswith('\\e'):
            res.append(char_to_keyinfo('\033', control, meta, shift))
            control = meta = shift = False
            keyseq = keyseq[2:]
        elif len(keyseq) >= 1:
            res.append(char_to_keyinfo(keyseq[0], control, meta, shift))
            control = meta = shift = False
            keyseq = keyseq[1:]
        else:
            return res[0]

def make_keyinfo(keycode, state):
    control = False
    meta  =False
    shift = False
    return (control, meta, shift, keycode)


def make_KeyPress(char, state, keycode):

    shift = bool(int(state) & int(Shift))
    control = bool(int(state) & int(Control))
    meta = bool(int(state) & int(Alt))
    keyname = code2sym_map.get(keycode, "").lower()
    if control and meta: #equivalent to altgr so clear flags
        control = False
        meta = False    
    elif control:
        char = str(keycode)
    return KeyPress(char, shift, control, meta, keyname)




# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Jörgen Stenarson. <>
from __future__ import print_function, unicode_literals, absolute_import

import sys, unittest
sys.path.append ('../..')
#from pyreadline.modes.vi import *
#from pyreadline import keysyms
from pyreadline.lineeditor import lineobj
from pyreadline.lineeditor.history import LineHistory
import pyreadline.lineeditor.history as history

import pyreadline.logger
pyreadline.logger.sock_silent=False
from pyreadline.logger import log
#----------------------------------------------------------------------


#----------------------------------------------------------------------
RL=lineobj.ReadLineTextBuffer

class Test_prev_next_history(unittest.TestCase):
    t = "test text"

    def setUp(self):
        self.q = q = LineHistory()
        for x in ["aaaa", "aaba", "aaca", "akca", "bbb", "ako"]:
            q.add_history(RL(x))

    def test_previous_history (self):
        hist = self.q
        assert hist.history_cursor == 6
        l = RL("")
        hist.previous_history(l)
        assert l.get_line_text() == "ako"
        hist.previous_history(l)
        assert l.get_line_text() == "bbb"
        hist.previous_history(l)
        assert l.get_line_text() == "akca"
        hist.previous_history(l)
        assert l.get_line_text() == "aaca"
        hist.previous_history(l)
        assert l.get_line_text() == "aaba"
        hist.previous_history(l)
        assert l.get_line_text() == "aaaa"
        hist.previous_history(l)
        assert l.get_line_text() == "aaaa"

    def test_next_history (self):
        hist=self.q
        hist.beginning_of_history()
        assert hist.history_cursor==0
        l=RL("")
        hist.next_history(l)
        assert l.get_line_text()=="aaba"
        hist.next_history(l)
        assert l.get_line_text()=="aaca"
        hist.next_history(l)
        assert l.get_line_text()=="akca"
        hist.next_history(l)
        assert l.get_line_text()=="bbb"
        hist.next_history(l)
        assert l.get_line_text()=="ako"
        hist.next_history(l)
        assert l.get_line_text()=="ako"

class Test_prev_next_history(unittest.TestCase):
    t = "test text"

    def setUp(self):
        self.q = q = LineHistory()
        for x in ["aaaa","aaba","aaca","akca","bbb","ako"]:
            q.add_history(RL(x))
        
    def test_history_search_backward (self):
        q = LineHistory()
        for x in ["aaaa","aaba","aaca","    aacax","akca","bbb","ako"]:
            q.add_history(RL(x))
        a=RL("aa",point=2)
        for x in ["aaca","aaba","aaaa","aaaa"]:
            res=q.history_search_backward(a)
            assert res.get_line_text()==x
        
    def test_history_search_forward (self):
        q = LineHistory()
        for x in ["aaaa","aaba","aaca","    aacax","akca","bbb","ako"]:
            q.add_history(RL(x))
        q.beginning_of_history()
        a=RL("aa",point=2)
        for x in ["aaba","aaca","aaca"]:
            res=q.history_search_forward(a)
            assert res.get_line_text()==x

class Test_history_search_incr_fwd_backwd(unittest.TestCase):
    def setUp(self):
        self.q = q = LineHistory()
        for x in ["aaaa","aaba","aaca","akca","bbb","ako"]:
            q.add_history(RL(x))

    def test_backward_1(self):
        q = self.q
        self.assertEqual(q.reverse_search_history("b"), "bbb") 
        self.assertEqual(q.reverse_search_history("b"), "aaba") 
        self.assertEqual(q.reverse_search_history("bb"), "aaba") 

    def test_backward_2(self):
        q = self.q
        self.assertEqual(q.reverse_search_history("a"), "ako") 
        self.assertEqual(q.reverse_search_history("aa"), "aaca") 
        self.assertEqual(q.reverse_search_history("a"), "aaca") 
        self.assertEqual(q.reverse_search_history("ab"), "aaba") 


    def test_forward_1(self):
        q = self.q
        self.assertEqual(q.forward_search_history("a"), "ako") 

    def test_forward_2(self):
        q = self.q
        q.history_cursor = 0
        self.assertEqual(q.forward_search_history("a"), "aaaa") 
        self.assertEqual(q.forward_search_history("a"), "aaba") 
        self.assertEqual(q.forward_search_history("ak"), "akca") 
        self.assertEqual(q.forward_search_history("akl"), "akca") 
        self.assertEqual(q.forward_search_history("ak"), "akca") 
        self.assertEqual(q.forward_search_history("ako"), "ako") 
        
class Test_empty_history_search_incr_fwd_backwd(unittest.TestCase):
    def setUp(self):
        self.q = q = LineHistory()

    def test_backward_1(self):
        q = self.q
        self.assertEqual(q.reverse_search_history("b"), "") 

    def test_forward_1(self):
        q = self.q
        self.assertEqual(q.forward_search_history("a"), "") 


#----------------------------------------------------------------------
# utility functions

#----------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()

    l=lineobj.ReadLineTextBuffer("First Second Third")


# -*- coding: utf-8 -*-
#*****************************************************************************
#       Copyright (C) 2003-2006 Gary Bishop.
#       Copyright (C) 2006  Jorgen Stenarson. <jorgen.stenarson@bostream.nu>
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from __future__ import print_function, unicode_literals, absolute_import
from . import winconstants as c32
from   pyreadline.logger import log
from ctypes import windll
import ctypes
# table for translating virtual keys to X windows key symbols

from .common import KeyPress

code2sym_map = {c32.VK_CANCEL:     'cancel',
                c32.VK_BACK:       'backspace',
                c32.VK_TAB:        'tab',
                c32.VK_CLEAR:      'clear',
                c32.VK_RETURN:     'return',
                c32.VK_SHIFT:      'shift_l',
                c32.VK_CONTROL:    'control_l',
                c32.VK_MENU:       'alt_l',
                c32.VK_PAUSE:      'pause',
                c32.VK_CAPITAL:    'caps_lock',
                c32.VK_ESCAPE:     'escape',
                c32.VK_SPACE:      'space',
                c32.VK_PRIOR:      'prior',
                c32.VK_NEXT:       'next',
                c32.VK_END:        'end',
                c32.VK_HOME:       'home',
                c32.VK_LEFT:       'left',
                c32.VK_UP:         'up',
                c32.VK_RIGHT:      'right',
                c32.VK_DOWN:       'down',
                c32.VK_SELECT:     'select',
                c32.VK_PRINT:      'print',
                c32.VK_EXECUTE:    'execute',
                c32.VK_SNAPSHOT:   'snapshot',
                c32.VK_INSERT:     'insert',
                c32.VK_DELETE:     'delete',
                c32.VK_HELP:       'help',
                c32.VK_F1:         'f1',
                c32.VK_F2:         'f2',
                c32.VK_F3:         'f3',
                c32.VK_F4:         'f4',
                c32.VK_F5:         'f5',
                c32.VK_F6:         'f6',
                c32.VK_F7:         'f7',
                c32.VK_F8:         'f8',
                c32.VK_F9:         'f9',
                c32.VK_F10:        'f10',
                c32.VK_F11:        'f11',
                c32.VK_F12:        'f12',
                c32.VK_F13:        'f13',
                c32.VK_F14:        'f14',
                c32.VK_F15:        'f15',
                c32.VK_F16:        'f16',
                c32.VK_F17:        'f17',
                c32.VK_F18:        'f18',
                c32.VK_F19:        'f19',
                c32.VK_F20:        'f20',
                c32.VK_F21:        'f21',
                c32.VK_F22:        'f22',
                c32.VK_F23:        'f23',
                c32.VK_F24:        'f24',
                c32.VK_NUMLOCK:    'num_lock,',
                c32.VK_SCROLL:     'scroll_lock',
                c32.VK_APPS:       'vk_apps',
                c32.VK_PROCESSKEY: 'vk_processkey',
                c32.VK_ATTN:       'vk_attn',
                c32.VK_CRSEL:      'vk_crsel',
                c32.VK_EXSEL:      'vk_exsel',
                c32.VK_EREOF:      'vk_ereof',
                c32.VK_PLAY:       'vk_play',
                c32.VK_ZOOM:       'vk_zoom',
                c32.VK_NONAME:     'vk_noname',
                c32.VK_PA1:        'vk_pa1',
                c32.VK_OEM_CLEAR:  'vk_oem_clear',
                c32.VK_NUMPAD0:    'numpad0',
                c32.VK_NUMPAD1:    'numpad1',
                c32.VK_NUMPAD2:    'numpad2',
                c32.VK_NUMPAD3:    'numpad3',
                c32.VK_NUMPAD4:    'numpad4',
                c32.VK_NUMPAD5:    'numpad5',
                c32.VK_NUMPAD6:    'numpad6',
                c32.VK_NUMPAD7:    'numpad7',
                c32.VK_NUMPAD8:    'numpad8',
                c32.VK_NUMPAD9:    'numpad9',
                c32.VK_DIVIDE:     'divide',
                c32.VK_MULTIPLY:   'multiply',
                c32.VK_ADD:        'add',
                c32.VK_SUBTRACT:   'subtract',
                c32.VK_DECIMAL:    'vk_decimal'
               }

VkKeyScan = windll.user32.VkKeyScanA

def char_to_keyinfo(char, control=False, meta=False, shift=False):
    k=KeyPress()
    vk = VkKeyScan(ord(char))
    if vk & 0xffff == 0xffff:
        print('VkKeyScan("%s") = %x' % (char, vk))
        raise ValueError('bad key')
    if vk & 0x100:
        k.shift = True
    if vk & 0x200:
        k.control = True
    if vk & 0x400:
        k.meta = True
    k.char=chr(vk & 0xff)
    return k

def make_KeyPress(char, state, keycode):
    control = (state & (4+8)) != 0
    meta = (state & (1+2)) != 0
    shift = (state & 0x10) != 0
    if control and not meta:#Matches ctrl- chords should pass keycode as char
        char = chr(keycode)
    elif control and meta:  #Matches alt gr and should just pass on char
        control = False
        meta = False
    try:
        keyname=code2sym_map[keycode]
    except KeyError:
        keyname = ""
    out = KeyPress(char, shift, control, meta, keyname)
    return out

if __name__ == "__main__":
    import startup
    


from __future__ import print_function, unicode_literals, absolute_import

class baseconsole(object):
    def __init__(self):
        pass
        
    def bell(self):
        raise NotImplementedError

    def pos(self, x=None, y=None):
        '''Move or query the window cursor.'''
        raise NotImplementedError

    def size(self):
        raise NotImplementedError

    def rectangle(self, rect, attr=None, fill=' '):
        '''Fill Rectangle.'''
        raise NotImplementedError

    def write_scrolling(self, text, attr=None):
        '''write text at current cursor position while watching for scrolling.

        If the window scrolls because you are at the bottom of the screen
        buffer, all positions that you are storing will be shifted by the
        scroll amount. For example, I remember the cursor position of the
        prompt so that I can redraw the line but if the window scrolls,
        the remembered position is off.

        This variant of write tries to keep track of the cursor position
        so that it will know when the screen buffer is scrolled. It
        returns the number of lines that the buffer scrolled.

        '''
        raise NotImplementedError
    
    def getkeypress(self):
        '''Return next key press event from the queue, ignoring others.'''
        raise NotImplementedError
        
    def write(self, text):
        raise NotImplementedError
    
    def page(self, attr=None, fill=' '):
        '''Fill the entire screen.'''
        raise NotImplementedError

    def isatty(self):
        return True

    def flush(self):
        pass

    